from decimal import Decimal
from enum import IntEnum
from typing import Union, List, Optional

from .operation import Operation
from ..asset import Asset
from ..keypair import Keypair
from ..strkey import StrKey
from ..utils import pack_xdr_array
from ..xdr import Xdr
from ..exceptions import ValueError
from .utils import check_amount

__all__ = ["ClaimPredicate", "Claimant", "CreateClaimableBalance"]


class ClaimPredicateType(IntEnum):
    """Currently supported claim predicate types.
    """

    CLAIM_PREDICATE_UNCONDITIONAL = 0
    CLAIM_PREDICATE_AND = 1
    CLAIM_PREDICATE_OR = 2
    CLAIM_PREDICATE_NOT = 3
    CLAIM_PREDICATE_BEFORE_ABSOLUTE_TIME = 4
    CLAIM_PREDICATE_BEFORE_RELATIVE_TIME = 5


class ClaimPredicateGroup:
    """Used to assemble the left and right values for and_predicates and or_predicates.

    :param left: The ClaimPredicate.
    :param right: The ClaimPredicate.
    """

    def __init__(self, left: "ClaimPredicate", right: "ClaimPredicate") -> None:
        self.left = left
        self.right = right


class ClaimPredicate:
    """The :class:`ClaimPredicate` object, which represents a ClaimPredicate on Stellar's network.

    **We do not recommend that you build it through the constructor, please use the helper function.**

    :param claim_predicate_type: Type of ClaimPredicate.
    :param and_predicates: The ClaimPredicates.
    :param or_predicates: The ClaimPredicates.
    :param not_predicate: The ClaimPredicate.
    :param abs_before: Unix epoch.
    :param rel_before: seconds since closeTime of the ledger in which the ClaimableBalanceEntry was created.
    """

    def __init__(
        self,
        claim_predicate_type: ClaimPredicateType,
        and_predicates: Optional[ClaimPredicateGroup],
        or_predicates: Optional[ClaimPredicateGroup],
        not_predicate: Optional["ClaimPredicate"],
        abs_before: Optional[int],
        rel_before: Optional[int],
    ) -> None:
        self.claim_predicate_type = claim_predicate_type
        self.and_predicates = and_predicates
        self.or_predicates = or_predicates
        self.not_predicate = not_predicate
        self.abs_before = abs_before
        self.rel_before = rel_before

    @classmethod
    def predicate_and(
        cls, left: "ClaimPredicate", right: "ClaimPredicate"
    ) -> "ClaimPredicate":
        """Returns an `and` claim predicate

        :param left: a ClaimPredicate.
        :param right: a ClaimPredicate.
        :return: an `and` claim predicate.
        """
        return cls(
            claim_predicate_type=ClaimPredicateType.CLAIM_PREDICATE_AND,
            and_predicates=ClaimPredicateGroup(left, right),
            or_predicates=None,
            not_predicate=None,
            abs_before=None,
            rel_before=None,
        )

    @classmethod
    def predicate_or(
        cls, left: "ClaimPredicate", right: "ClaimPredicate"
    ) -> "ClaimPredicate":
        """Returns an `or` claim predicate

        :param left: a ClaimPredicate.
        :param right: a ClaimPredicate.
        :return: an `or` claim predicate.
        """
        return cls(
            claim_predicate_type=ClaimPredicateType.CLAIM_PREDICATE_OR,
            and_predicates=None,
            or_predicates=ClaimPredicateGroup(left, right),
            not_predicate=None,
            abs_before=None,
            rel_before=None,
        )

    @classmethod
    def predicate_not(cls, predicate: "ClaimPredicate") -> "ClaimPredicate":
        """Returns a `not` claim predicate.

        :param predicate: a ClaimPredicate.
        :return: a `not` claim predicate.
        """
        return cls(
            claim_predicate_type=ClaimPredicateType.CLAIM_PREDICATE_NOT,
            and_predicates=None,
            or_predicates=None,
            not_predicate=predicate,
            abs_before=None,
            rel_before=None,
        )

    @classmethod
    def predicate_before_absolute_time(cls, abs_before: int) -> "ClaimPredicate":
        """Returns a `before_absolute_time` claim predicate.

        This predicate will be fulfilled if the closing time of the ledger that includes
        the :class:`CreateClaimableBalance` operation is less than this (absolute) Unix timestamp.

        :param abs_before: Unix epoch.
        :return: a `before_absolute_time` claim predicate.
        """
        return cls(
            claim_predicate_type=ClaimPredicateType.CLAIM_PREDICATE_BEFORE_ABSOLUTE_TIME,
            and_predicates=None,
            or_predicates=None,
            not_predicate=None,
            abs_before=abs_before,
            rel_before=None,
        )

    @classmethod
    def predicate_before_relative_time(cls, seconds: int) -> "ClaimPredicate":
        """Returns a `before_relative_time` claim predicate.

        This predicate will be fulfilled if the closing time of the ledger that
        includes the :class:`CreateClaimableBalance` operation plus this relative time delta (in seconds)
        is less than the current time.

        :param seconds: seconds since closeTime of the ledger in which the ClaimableBalanceEntry was created.
        :return: a `before_relative_time` claim predicate.
        """
        return cls(
            claim_predicate_type=ClaimPredicateType.CLAIM_PREDICATE_BEFORE_RELATIVE_TIME,
            and_predicates=None,
            or_predicates=None,
            not_predicate=None,
            abs_before=None,
            rel_before=seconds,
        )

    @classmethod
    def predicate_unconditional(cls) -> "ClaimPredicate":
        """Returns an unconditional claim predicate.

        :return: an unconditional claim predicate.
        """
        return cls(
            claim_predicate_type=ClaimPredicateType.CLAIM_PREDICATE_UNCONDITIONAL,
            and_predicates=None,
            or_predicates=None,
            not_predicate=None,
            abs_before=None,
            rel_before=None,
        )

    def to_xdr_object(self) -> Xdr.nullclass:
        data = Xdr.nullclass()
        if (
            self.claim_predicate_type
            == ClaimPredicateType.CLAIM_PREDICATE_UNCONDITIONAL
        ):
            data.type = Xdr.const.CLAIM_PREDICATE_UNCONDITIONAL
            return data
        elif (
            self.claim_predicate_type
            == ClaimPredicateType.CLAIM_PREDICATE_BEFORE_ABSOLUTE_TIME
        ):
            data.type = Xdr.const.CLAIM_PREDICATE_BEFORE_ABSOLUTE_TIME
            data.absBefore = self.abs_before
            return data
        elif (
            self.claim_predicate_type
            == ClaimPredicateType.CLAIM_PREDICATE_BEFORE_RELATIVE_TIME
        ):
            data.type = Xdr.const.CLAIM_PREDICATE_BEFORE_RELATIVE_TIME
            data.relBefore = self.rel_before
            return data
        elif self.claim_predicate_type == ClaimPredicateType.CLAIM_PREDICATE_NOT:
            data.type = Xdr.const.CLAIM_PREDICATE_NOT
            data.notPredicate = pack_xdr_array(self.not_predicate.to_xdr_object())
            return data
        elif self.claim_predicate_type == ClaimPredicateType.CLAIM_PREDICATE_AND:
            data.type = Xdr.const.CLAIM_PREDICATE_AND
            data.andPredicates = [
                self.and_predicates.left.to_xdr_object(),
                self.and_predicates.right.to_xdr_object(),
            ]
            return data
        elif self.claim_predicate_type == ClaimPredicateType.CLAIM_PREDICATE_OR:
            data.type = Xdr.const.CLAIM_PREDICATE_OR
            data.orPredicates = [
                self.or_predicates.left.to_xdr_object(),
                self.or_predicates.right.to_xdr_object(),
            ]
            return data
        else:
            raise ValueError(
                f"{self.claim_predicate_type} is not a valid ClaimPredicateType."
            )

    @classmethod
    def from_xdr_object(cls, xdr_object: Xdr.nullclass) -> "ClaimPredicate":
        claim_predicate_type = ClaimPredicateType(xdr_object.type)
        if claim_predicate_type == ClaimPredicateType.CLAIM_PREDICATE_UNCONDITIONAL:
            return cls.predicate_unconditional()
        elif (
            claim_predicate_type
            == ClaimPredicateType.CLAIM_PREDICATE_BEFORE_ABSOLUTE_TIME
        ):
            abs_before = xdr_object.absBefore
            return cls.predicate_before_absolute_time(abs_before)
        elif (
            claim_predicate_type
            == ClaimPredicateType.CLAIM_PREDICATE_BEFORE_RELATIVE_TIME
        ):
            rel_before = xdr_object.relBefore
            return cls.predicate_before_relative_time(rel_before)
        elif claim_predicate_type == ClaimPredicateType.CLAIM_PREDICATE_NOT:
            not_pedicate = xdr_object.notPredicate
            pedicate = cls.from_xdr_object(not_pedicate[0])
            return cls.predicate_not(pedicate)
        elif claim_predicate_type == ClaimPredicateType.CLAIM_PREDICATE_AND:
            and_predicates = xdr_object.andPredicates
            left = cls.from_xdr_object(and_predicates[0])
            right = cls.from_xdr_object(and_predicates[1])
            return cls.predicate_and(left, right)
        elif claim_predicate_type == ClaimPredicateType.CLAIM_PREDICATE_OR:
            or_predicates = xdr_object.orPredicates
            left = cls.from_xdr_object(or_predicates[0])
            right = cls.from_xdr_object(or_predicates[1])
            return cls.predicate_or(left, right)
        else:
            raise ValueError(
                "{} is an unsupported ClaimPredicateType."
            )  # pragma: no cover


class Claimant:
    """The :class:`Claimant` object represents a claimable balance claimant.

    :param destination: The destination account ID.
    :param predicate: The claim predicate. It is optional, it defaults to unconditional if none is specified.
    """

    def __init__(self, destination: str, predicate: ClaimPredicate = None) -> None:
        self.destination = destination
        if predicate is None:
            predicate = ClaimPredicate.predicate_unconditional()
        self.predicate: ClaimPredicate = predicate

    def to_xdr_object(self) -> Xdr.nullclass:
        v0 = Xdr.nullclass()
        v0.destination = Keypair.from_public_key(self.destination).xdr_account_id()
        v0.predicate = self.predicate.to_xdr_object()
        data = Xdr.nullclass()

        data.type = Xdr.const.CLAIMANT_TYPE_V0
        data.v0 = v0
        return data

    @classmethod
    def from_xdr_object(cls, xdr_object: Xdr.nullclass) -> "Claimant":
        destination = StrKey.encode_ed25519_public_key(
            xdr_object.v0.destination.ed25519
        )
        predicate = ClaimPredicate.from_xdr_object(xdr_object.v0.predicate)
        return cls(destination=destination, predicate=predicate)


class CreateClaimableBalance(Operation):
    """The :class:`CreateClaimableBalance` object, which represents a CreateClaimableBalance
    operation on Stellar's network.

    Creates a ClaimableBalanceEntry. See `Claimable Balance
    <https://developers.stellar.org/docs/glossary/claimable-balance/>_` for more information on parameters and usage.

    See `Create Claimable Balance
    <https://developers.stellar.org/docs/start/list-of-operations/#create-claimable-balance>_`.

    Threshold: Medium

    :param asset: The asset for the claimable balance.
    :param amount: the amount of the asset.
    :param claimants: A list of Claimants.
    :param source: The source account (defaults to transaction source).
    """

    def __init__(
        self,
        asset: Asset,
        amount: Union[str, Decimal],
        claimants: List[Claimant],
        source: str = None,
    ) -> None:
        super().__init__(source)
        check_amount(amount)
        self.asset = asset
        self.amount = amount
        self.claimants = claimants
        self.source = source

    @classmethod
    def type_code(cls) -> int:
        return Xdr.const.CREATE_CLAIMABLE_BALANCE

    def _to_operation_body(self) -> Xdr.nullclass:
        body = Xdr.nullclass()
        body.type = Xdr.const.CREATE_CLAIMABLE_BALANCE
        asset = self.asset.to_xdr_object()
        amount = Operation.to_xdr_amount(self.amount)
        claimants = [claimant.to_xdr_object() for claimant in self.claimants]
        create_claimable_balance_op = Xdr.types.CreateClaimableBalanceOp(
            asset=asset, amount=amount, claimants=claimants
        )
        body.createClaimableBalanceOp = create_claimable_balance_op
        return body

    @classmethod
    def from_xdr_object(
        cls, operation_xdr_object: Xdr.types.Operation
    ) -> "CreateClaimableBalance":
        """Creates a :class:`CreateClaimableBalance` object from an XDR Operation
        object.
        """
        source = Operation.get_source_from_xdr_obj(operation_xdr_object)
        asset = Asset.from_xdr_object(
            operation_xdr_object.body.createClaimableBalanceOp.asset
        )
        amount = Operation.from_xdr_amount(
            operation_xdr_object.body.createClaimableBalanceOp.amount
        )
        claimants = []
        for (
            claimant_xdr_obj
        ) in operation_xdr_object.body.createClaimableBalanceOp.claimants:
            claimants.append(Claimant.from_xdr_object(claimant_xdr_obj))
        op = cls(asset=asset, amount=amount, claimants=claimants, source=source)
        op._source_muxed = Operation.get_source_muxed_from_xdr_obj(operation_xdr_object)
        return op



from decimal import Decimal
from typing import Union

from .operation import Operation
from .utils import check_amount, check_price
from ..asset import Asset
from ..price import Price
from ..xdr import Xdr


class CreatePassiveSellOffer(Operation):
    """The :class:`CreatePassiveSellOffer` object, which represents a
    CreatePassiveSellOffer operation on Stellar's network.

    A passive sell offer is an offer that does not act on and take a reverse offer
    of equal price. Instead, they only take offers of lesser price. For
    example, if an offer exists to buy 5 BTC for 30 XLM, and you make a passive
    sell offer to buy 30 XLM for 5 BTC, your passive sell offer does not take the first
    offer.

    Note that regular offers made later than your passive sell offer can act on and
    take your passive sell offer, even if the regular offer is of the same price as
    your passive sell offer.

    Passive sell offers allow market makers to have zero spread. If you want to
    trade EUR for USD at 1:1 price and USD for EUR also at 1:1, you can create
    two passive sell offers so the two offers don't immediately act on each other.

    Once the passive sell offer is created, you can manage it like any other offer
    using the manage offer operation - see :class:`ManageOffer` for more
    details.

    :param selling: What you're selling.
    :param buying: What you're buying.
    :param amount: The total amount you're selling. If 0,
        deletes the offer.
    :param price: Price of 1 unit of `selling` in
        terms of `buying`.
    :param source: The source account (defaults to transaction source).

    """

    def __init__(
        self,
        selling: Asset,
        buying: Asset,
        amount: Union[str, Decimal],
        price: Union[Price, str, Decimal],
        source: str = None,
    ) -> None:
        super().__init__(source)
        check_amount(amount)
        check_price(price)
        self.selling: Asset = selling
        self.buying: Asset = buying
        self.amount: Union[str, Decimal] = amount
        self.price: Union[Price, str, Decimal] = price

    @classmethod
    def type_code(cls) -> int:
        return Xdr.const.CREATE_PASSIVE_SELL_OFFER

    def _to_operation_body(self) -> Xdr.nullclass:
        selling = self.selling.to_xdr_object()
        buying = self.buying.to_xdr_object()

        if isinstance(self.price, Price):
            price_fraction = self.price
        else:
            price_fraction = Price.from_raw_price(self.price)

        price = price_fraction.to_xdr_object()

        amount = Operation.to_xdr_amount(self.amount)

        create_passive_sell_offer_op = Xdr.types.CreatePassiveSellOfferOp(
            selling, buying, amount, price
        )
        body = Xdr.nullclass()
        body.type = Xdr.const.CREATE_PASSIVE_SELL_OFFER
        body.createPassiveSellOfferOp = create_passive_sell_offer_op
        return body

    @classmethod
    def from_xdr_object(
        cls, operation_xdr_object: Xdr.types.Operation
    ) -> "CreatePassiveSellOffer":
        """Creates a :class:`CreatePassiveSellOffer` object from an XDR Operation object.

        """
        source = Operation.get_source_from_xdr_obj(operation_xdr_object)
        selling = Asset.from_xdr_object(
            operation_xdr_object.body.createPassiveSellOfferOp.selling
        )
        buying = Asset.from_xdr_object(
            operation_xdr_object.body.createPassiveSellOfferOp.buying
        )
        amount = Operation.from_xdr_amount(
            operation_xdr_object.body.createPassiveSellOfferOp.amount
        )
        price = Price.from_xdr_object(
            operation_xdr_object.body.createPassiveSellOfferOp.price
        )

        op = cls(
            source=source, selling=selling, buying=buying, amount=amount, price=price
        )
        op._source_muxed = Operation.get_source_muxed_from_xdr_obj(operation_xdr_object)
        return op



from typing import Union

from ..asset import Asset
from ..call_builder.base_call_builder import BaseCallBuilder
from ..client.base_async_client import BaseAsyncClient
from ..client.base_sync_client import BaseSyncClient


class OrderbookCallBuilder(BaseCallBuilder):
    """ Creates a new :class:`OrderbookCallBuilder` pointed to server defined by horizon_url.
    Do not create this object directly, use :func:`stellar_sdk.server.Server.orderbook`.

    See `Orderbook Details <https://www.stellar.org/developers/horizon/reference/endpoints/orderbook-details.html>`_

    :param horizon_url: Horizon server URL.
    :param client: The client instance used to send request.
    :param selling: Asset being sold
    :param buying: Asset being bought
    """

    def __init__(
        self,
        horizon_url: str,
        client: Union[BaseAsyncClient, BaseSyncClient],
        selling: Asset,
        buying: Asset,
    ) -> None:
        super().__init__(horizon_url, client)
        self.endpoint: str = "order_book"
        params = {
            "selling_asset_type": selling.type,
            "selling_asset_code": None if selling.is_native() else selling.code,
            "selling_asset_issuer": selling.issuer,
            "buying_asset_type": buying.type,
            "buying_asset_code": None if buying.is_native() else buying.code,
            "buying_asset_issuer": buying.issuer,
        }
        self._add_query_params(params)



from .operation import Operation
from ..xdr import Xdr


class EndSponsoringFutureReserves(Operation):
    """The :class:`EndSponsoringFutureReserves` object, which represents a EndSponsoringFutureReserves
    operation on Stellar's network.

    Terminates the current is-sponsoring-future-reserves-for relationship in which the source account is sponsored.
    See `Sponsored Reserves <https://developers.stellar.org/docs/glossary/sponsored-reserves/>_` for more information.

    See `End Sponsoring Future Reserves
    <https://developers.stellar.org/docs/start/list-of-operations/#end-sponsoring-future-reserves>_`.

    Threshold: Medium

    :param source: The source account (defaults to transaction source).
    """

    def __init__(self, source: str = None) -> None:
        super().__init__(source)

    @classmethod
    def type_code(cls) -> int:
        return Xdr.const.END_SPONSORING_FUTURE_RESERVES

    def _to_operation_body(self) -> Xdr.nullclass:
        body = Xdr.nullclass()
        body.type = Xdr.const.END_SPONSORING_FUTURE_RESERVES
        return body

    @classmethod
    def from_xdr_object(
        cls, operation_xdr_object: Xdr.types.Operation
    ) -> "EndSponsoringFutureReserves":
        """Creates a :class:`EndSponsoringFutureReserves` object from an XDR Operation
        object.
        """
        source = Operation.get_source_from_xdr_obj(operation_xdr_object)
        op = cls(source=source,)
        op._source_muxed = Operation.get_source_muxed_from_xdr_obj(operation_xdr_object)
        return op



from typing import Union

from ..call_builder.base_call_builder import BaseCallBuilder
from ..client.base_async_client import BaseAsyncClient
from ..client.base_sync_client import BaseSyncClient


class RootCallBuilder(BaseCallBuilder):
    """ Creates a new :class:`RootCallBuilder` pointed to server defined by horizon_url.
    Do not create this object directly, use :func:`stellar_sdk.server.Server.root`.

    :param horizon_url: Horizon server URL.
    :param client: The client instance used to send request.
    """

    def __init__(
        self, horizon_url: str, client: Union[BaseAsyncClient, BaseSyncClient]
    ) -> None:
        super().__init__(horizon_url, client)



from decimal import Decimal
from typing import List, Union, Optional

from .operation import Operation
from .utils import check_amount, check_ed25519_public_key
from ..asset import Asset
from ..keypair import Keypair
from ..utils import parse_ed25519_account_id_from_muxed_account_xdr_object
from ..xdr import Xdr


class PathPaymentStrictReceive(Operation):
    """The :class:`PathPaymentStrictReceive` object, which represents a PathPaymentStrictReceive
    operation on Stellar's network.

    Sends an amount in a specific asset to a destination account through a path
    of offers. This allows the asset sent (e.g. 450 XLM) to be different from
    the asset received (e.g. 6 BTC).

    Threshold: Medium

    :param destination: The destination account to send to.
    :param send_asset: The asset to pay with.
    :param send_max: The maximum amount of send_asset to send.
    :param dest_asset: The asset the destination will receive.
    :param dest_amount: The amount the destination receives.
    :param path: A list of Asset objects to use as the path.
    :param source: The source account for the payment. Defaults to the
        transaction's source account.
    """

    def __init__(
        self,
        destination: str,
        send_asset: Asset,
        send_max: Union[str, Decimal],
        dest_asset: Asset,
        dest_amount: Union[str, Decimal],
        path: List[Asset],
        source: str = None,
    ) -> None:
        super().__init__(source)
        check_amount(send_max)
        check_amount(dest_amount)
        check_ed25519_public_key(destination)
        self._destination: str = destination
        self._destination_muxed: Optional[Xdr.types.MuxedAccount] = None
        self.send_asset: Asset = send_asset
        self.send_max: Union[str, Decimal] = send_max
        self.dest_asset: Asset = dest_asset
        self.dest_amount: Union[str, Decimal] = dest_amount
        self.path: List[Asset] = path  # a list of paths/assets

    @property
    def destination(self) -> str:
        return self._destination

    @destination.setter
    def destination(self, value: str):
        check_ed25519_public_key(value)
        self._destination_muxed = None
        self._destination = value

    @classmethod
    def type_code(cls) -> int:
        return Xdr.const.PATH_PAYMENT_STRICT_RECEIVE

    def _to_operation_body(self) -> Xdr.nullclass:
        if self._destination_muxed is not None:
            destination = self._destination_muxed
        else:
            destination = Keypair.from_public_key(self._destination).xdr_muxed_account()
        send_asset = self.send_asset.to_xdr_object()
        dest_asset = self.dest_asset.to_xdr_object()
        path = [asset.to_xdr_object() for asset in self.path]

        path_payment_strict_receive_op = Xdr.types.PathPaymentStrictReceiveOp(
            send_asset,
            Operation.to_xdr_amount(self.send_max),
            destination,
            dest_asset,
            Operation.to_xdr_amount(self.dest_amount),
            path,
        )
        body = Xdr.nullclass()
        body.type = Xdr.const.PATH_PAYMENT_STRICT_RECEIVE
        body.pathPaymentStrictReceiveOp = path_payment_strict_receive_op
        return body

    @classmethod
    def from_xdr_object(
        cls, operation_xdr_object: Xdr.types.Operation
    ) -> "PathPaymentStrictReceive":
        """Creates a :class:`PathPaymentStrictReceive` object from an XDR Operation
        object.

        """
        source = Operation.get_source_from_xdr_obj(operation_xdr_object)
        destination = parse_ed25519_account_id_from_muxed_account_xdr_object(
            operation_xdr_object.body.pathPaymentStrictReceiveOp.destination
        )

        send_asset = Asset.from_xdr_object(
            operation_xdr_object.body.pathPaymentStrictReceiveOp.sendAsset
        )
        dest_asset = Asset.from_xdr_object(
            operation_xdr_object.body.pathPaymentStrictReceiveOp.destAsset
        )
        send_max = Operation.from_xdr_amount(
            operation_xdr_object.body.pathPaymentStrictReceiveOp.sendMax
        )
        dest_amount = Operation.from_xdr_amount(
            operation_xdr_object.body.pathPaymentStrictReceiveOp.destAmount
        )

        path = []
        if operation_xdr_object.body.pathPaymentStrictReceiveOp.path:
            for x in operation_xdr_object.body.pathPaymentStrictReceiveOp.path:
                path.append(Asset.from_xdr_object(x))

        op = cls(
            source=source,
            destination=destination,
            send_asset=send_asset,
            send_max=send_max,
            dest_asset=dest_asset,
            dest_amount=dest_amount,
            path=path,
        )
        op._destination_muxed = (
            operation_xdr_object.body.pathPaymentStrictReceiveOp.destination
        )
        op._source_muxed = Operation.get_source_muxed_from_xdr_obj(operation_xdr_object)
        return op



"""
SEP: 0001
Title: stellar.toml
Author: stellar.org
Status: Active
Created: 2017-10-30
Updated: 2019-06-12
Version: 2.1.0
"""
from typing import Union, Dict, Any, Coroutine

import toml

from .exceptions import StellarTomlNotFoundError
from ..client.base_async_client import BaseAsyncClient
from ..client.base_sync_client import BaseSyncClient
from ..client.requests_client import RequestsClient
from ..client.response import Response


def fetch_stellar_toml(
    domain: str,
    client: Union[BaseAsyncClient, BaseSyncClient] = None,
    use_http: bool = False,
) -> Union[Coroutine[Any, Any, Dict[str, Any]], Dict[str, Any]]:
    """Retrieve the stellar.toml file from a given domain.

    Retrieve the stellar.toml file for information about interacting with
    Stellar's federation protocol for a given Stellar Anchor (specified by a
    domain).

    :param domain: The domain the .toml file is hosted at.
    :param use_http: Specifies whether the request should go over plain HTTP vs HTTPS.
        Note it is recommend that you *always* use HTTPS.
    :param client: Http Client used to send the request.
    :return: The stellar.toml file as a an object via :func:`toml.loads`.
    :raises: :exc:`StellarTomlNotFoundError <stellar_sdk.sep.exceptions.StellarTomlNotFoundError>`:
        if the Stellar toml file could not not be found.
    """
    if not client:
        client = RequestsClient()

    toml_link = "/.well-known/stellar.toml"
    protocol = "https://"
    if use_http:
        protocol = "http://"
    url = protocol + domain + toml_link

    if isinstance(client, BaseAsyncClient):
        return __fetch_async(url, client)
    elif isinstance(client, BaseSyncClient):
        return __fetch_sync(url, client)
    else:
        raise TypeError(
            "This `client` class should be an instance "
            "of `stellar_sdk.client.base_async_client.BaseAsyncClient` "
            "or `stellar_sdk.client.base_sync_client.BaseSyncClient`."
        )


async def __fetch_async(url: str, client: BaseAsyncClient) -> Dict[str, Any]:
    raw_resp = await client.get(url)
    return __handle_raw_response(raw_resp)


def __fetch_sync(url: str, client: BaseSyncClient) -> Dict[str, Any]:
    raw_resp = client.get(url)
    return __handle_raw_response(raw_resp)


def __handle_raw_response(raw_resp: Response) -> Dict[str, Any]:
    if raw_resp.status_code == 404:
        raise StellarTomlNotFoundError
    resp = raw_resp.text
    return toml.loads(resp)



from typing import Optional

from .operation import Operation
from .utils import check_ed25519_public_key
from ..keypair import Keypair
from ..utils import parse_ed25519_account_id_from_muxed_account_xdr_object
from ..xdr import Xdr


class AccountMerge(Operation):
    """The :class:`AccountMerge` object, which represents a
    AccountMerge operation on Stellar's network.

    Transfers the native balance (the amount of XLM an account holds) to
    another account and removes the source account from the ledger.

    Threshold: High

    :param destination: Destination to merge the source account into.
    :param source: The source account (defaults to transaction source).

    """

    def __init__(self, destination: str, source: str = None,) -> None:
        super().__init__(source)
        check_ed25519_public_key(destination)
        self._destination: str = destination
        self._destination_muxed: Optional[Xdr.types.MuxedAccount] = None

    @property
    def destination(self) -> str:
        return self._destination

    @destination.setter
    def destination(self, value: str):
        check_ed25519_public_key(value)
        self._destination_muxed = None
        self._destination = value

    @classmethod
    def type_code(cls) -> int:
        return Xdr.const.ACCOUNT_MERGE

    def _to_operation_body(self) -> Xdr.nullclass:
        if self._destination_muxed is not None:
            destination = self._destination_muxed
        else:
            destination = Keypair.from_public_key(self._destination).xdr_muxed_account()
        body = Xdr.nullclass()
        body.type = Xdr.const.ACCOUNT_MERGE
        body.destination = destination
        return body

    @classmethod
    def from_xdr_object(
        cls, operation_xdr_object: Xdr.types.Operation
    ) -> "AccountMerge":
        """Creates a :class:`AccountMerge` object from an XDR Operation
        object.

        """
        source = Operation.get_source_from_xdr_obj(operation_xdr_object)
        destination = parse_ed25519_account_id_from_muxed_account_xdr_object(
            operation_xdr_object.body.destination
        )
        op = cls(source=source, destination=destination)
        op._destination_muxed = operation_xdr_object.body.destination
        op._source_muxed = Operation.get_source_muxed_from_xdr_obj(operation_xdr_object)
        return op



from typing import Union

from ..call_builder.base_call_builder import BaseCallBuilder
from ..client.base_async_client import BaseAsyncClient
from ..client.base_sync_client import BaseSyncClient


class LedgersCallBuilder(BaseCallBuilder):
    """ Creates a new :class:`LedgersCallBuilder` pointed to server defined by horizon_url.
    Do not create this object directly, use :func:`stellar_sdk.server.Server.ledgers`.

    See `All Ledgers <https://www.stellar.org/developers/horizon/reference/endpoints/ledgers-all.html>`_

    :param horizon_url: Horizon server URL.
    :param client: The client instance used to send request.
    """

    def __init__(
        self, horizon_url: str, client: Union[BaseAsyncClient, BaseSyncClient]
    ) -> None:
        super().__init__(horizon_url, client)
        self.endpoint: str = "ledgers"

    def ledger(self, sequence: Union[int, str]) -> "LedgersCallBuilder":
        """Provides information on a single ledger.

        See `Ledger Details <https://www.stellar.org/developers/horizon/reference/endpoints/ledgers-single.html>`_

        :param sequence: Ledger sequence
        :return: current LedgerCallBuilder instance
        """
        self.endpoint = f"ledgers/{sequence}"
        return self



from typing import Union

from ..call_builder.base_call_builder import BaseCallBuilder
from ..client.base_async_client import BaseAsyncClient
from ..client.base_sync_client import BaseSyncClient


class OperationsCallBuilder(BaseCallBuilder):
    """ Creates a new :class:`OperationsCallBuilder` pointed to server defined by horizon_url.
    Do not create this object directly, use :func:`stellar_sdk.server.Server.operations`.

    See `All Operations <https://www.stellar.org/developers/horizon/reference/endpoints/operations-all.html>`_

    :param horizon_url: Horizon server URL.
    :param client: The client instance used to send request.
    """

    def __init__(
        self, horizon_url, client: Union[BaseAsyncClient, BaseSyncClient]
    ) -> None:
        super().__init__(horizon_url, client)
        self.endpoint: str = "operations"

    def operation(self, operation_id: Union[int, str]) -> "OperationsCallBuilder":
        """The operation details endpoint provides information on a single operation. The operation ID provided
        in the id argument specifies which operation to load.

        See `Operation Details <https://www.stellar.org/developers/horizon/reference/endpoints/operations-single.html>`_

        :param operation_id: Operation ID
        :return: this OperationCallBuilder instance
        """
        self.endpoint = f"operations/{operation_id}"
        return self

    def for_account(self, account_id: str) -> "OperationsCallBuilder":
        """This endpoint represents all operations that were included in valid transactions that
        affected a particular account.

        See `Operations for Account <https://www.stellar.org/developers/horizon/reference/endpoints/operations-for-account.html>`_

        :param account_id: Account ID
        :return: this OperationCallBuilder instance
        """
        self.endpoint: str = f"accounts/{account_id}/operations"
        return self

    def for_ledger(self, sequence: Union[int, str]) -> "OperationsCallBuilder":
        """This endpoint returns all operations that occurred in a given ledger.

        See `Operations for Ledger <https://www.stellar.org/developers/horizon/reference/endpoints/operations-for-ledger.html>`_

        :param sequence: Sequence ID
        :return: this OperationCallBuilder instance
        """
        self.endpoint: str = f"ledgers/{sequence}/operations"
        return self

    def for_transaction(self, transaction_hash: str) -> "OperationsCallBuilder":
        """This endpoint represents all operations that are part of a given transaction.

        See `Operations for Transaction <https://www.stellar.org/developers/horizon/reference/endpoints/operations-for-transaction.html>`_

        :param transaction_hash:
        :return: this OperationCallBuilder instance
        """
        self.endpoint: str = f"transactions/{transaction_hash}/operations"
        return self

    def include_failed(self, include_failed: bool) -> "OperationsCallBuilder":
        """Adds a parameter defining whether to include failed transactions. By default only
        operations of successful transactions are returned.

        :param include_failed: Set to `True` to include operations of failed transactions.
        :return: current OperationsCallBuilder instance
        """
        self._add_query_param("include_failed", include_failed)
        return self

    def join(self, join: str) -> "OperationsCallBuilder":
        """join represents `join` param in queries, currently only supports `transactions`

        :param join: join represents `join` param in queries, currently only supports `transactions`
        :return: current OperationsCallBuilder instance
        """
        self._add_query_param("join", join)
        return self



from abc import ABCMeta, abstractmethod
from typing import Dict, Any, Generator

from .response import Response


class BaseSyncClient(metaclass=ABCMeta):
    """This is an abstract class,
    and if you want to implement your own synchronous client, you **must** implement this class.

    """

    @abstractmethod
    def get(self, url: str, params: Dict[str, str] = None) -> Response:
        """Perform HTTP GET request.

        :param url: the request url
        :param params: the request params
        :return: the response from server
        :raise: :exc:`ConnectionError <stellar_sdk.exceptions.ConnectionError>`
        """
        pass

    @abstractmethod
    def post(self, url: str, data: Dict[str, str]) -> Response:
        """Perform HTTP POST request.

        :param url: the request url
        :param data: the data send to server
        :return: the response from server
        :raise: :exc:`ConnectionError <stellar_sdk.exceptions.ConnectionError>`
        """
        pass

    @abstractmethod
    def stream(
        self, url: str, params: Dict[str, str] = None
    ) -> Generator[Dict[str, Any], None, None]:
        """Creates an EventSource that listens for incoming messages from the server.

        See `Horizon Response Format <https://www.stellar.org/developers/horizon/reference/responses.html>`_

        See `MDN EventSource <https://developer.mozilla.org/en-US/docs/Web/API/EventSource>`_

        :param url: the request url
        :param params: the request params
        :return: a dict Generator for server response
        :raise: :exc:`ConnectionError <stellar_sdk.exceptions.ConnectionError>`
        """
        pass

    @abstractmethod
    def close(self):
        pass



#!/usr/bin/env python
from __future__ import print_function

import os
import sys
import time

import ply.lex as lex

# xdrgen.py for stellar
#
# rpcgen.py - A Python RPC protocol compiler
#
# Written by Fred Isaman <iisaman@citi.umich.edu>
# Copyright (C) 2004 University of Michigan, Center for
#                    Information Technology Integration
#
# Based on version written by Peter Astrand <peter@cendio.se>
# Copyright (C) 2001 Cendio Systems AB (http://www.cendio.se)
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#
# Note <something>_list means zero or more of <something>.
#
# TODO:
# Code generation for programs and procedures.
#

##########################################################################
#                                                                        #
# Code is based on the following updated sections of RFCs 1831 and 1832: #
#                                                                        #
##########################################################################
"""
6.2.  Lexical Notes

   (1) Comments begin with '/*' and terminate with '*/'.  (2) White
   space serves to separate items and is otherwise ignored.  (3) An
   identifier is a letter followed by an optional sequence of letters,
   digits or underbar ('_'). The case of identifiers is not ignored.
   (4) A decimal constant expresses a number in base 10, and is a
   sequence of one or more decimal digits, where the first digit is not
   a zero, and is optionally preceded by a minus-sign ('-').  (5) A
   hexadecimal constant expresses a number in base 16, and must be
   preceded by '0x', followed by one or hexadecimal digits ('A', 'B',
   'C', 'D', E', 'F', 'a', 'b', 'c', 'd', 'e', 'f', '0', '1', '2', '3',
   '4', '5', '6', '7', '8', '9'). (6) An octal constant expresses a
   number in base 8, always leads with digit 0, and is a sequence of
   one or more octal digits ('0', '1', '2', '3', '4', '5', '6', '7').


6.3.  Syntax Information

      declaration:
           type-specifier identifier
         | type-specifier identifier "[" value "]"
         | type-specifier identifier "<" [ value ] ">"
         | "opaque" identifier "[" value "]"
         | "opaque" identifier "<" [ value ] ">"
         | "string" identifier "<" [ value ] ">"
         | type-specifier "*" identifier
         | "void"

      value:
           constant
         | identifier

      constant:
         decimal-constant | hexadecimal-constant | octal-constant

      type-specifier:
           [ "unsigned" ] "int"
         | [ "unsigned" ] "hyper"
         | "float"
         | "double"
         | "quadruple"
         | "bool"
         | enum-type-spec
         | struct-type-spec
         | union-type-spec
         | identifier

      enum-type-spec:
         "enum" enum-body

      enum-body:
         "{"
            ( identifier "=" value )
            ( "," identifier "=" value )*
         "}"

      struct-type-spec:
         "struct" struct-body

      struct-body:
         "{"
            ( declaration ";" )
            ( declaration ";" )*
         "}"

      union-type-spec:
         "union" union-body

      union-body:
         "switch" "(" declaration ")" "{"
            case-spec
            case-spec *
            [ "default" ":" declaration ";" ]
         "}"

      case-spec:
        ( "case" value ":")
        ( "case" value ":") *
        declaration ";"

      constant-def:
         "const" identifier "=" constant ";"

      type-def:
           "typedef" declaration ";"
         | "enum" identifier enum-body ";"
         | "struct" identifier struct-body ";"
         | "union" identifier union-body ";"

      definition:
           type-def
         | constant-def

      specification:
           definition *


6.4.  Syntax Notes

   (1) The following are keywords and cannot be used as identifiers:
   "bool", "case", "const", "default", "double", "quadruple", "enum",
   "float", "hyper", "int", "opaque", "string", "struct", "switch",
   "typedef", "union", "unsigned" and "void".

   (2) Only unsigned constants may be used as size specifications for
   arrays.  If an identifier is used, it must have been declared
   previously as an unsigned constant in a "const" definition.

   (3) Constant and type identifiers within the scope of a specification
   are in the same name space and must be declared uniquely within this
   scope.

   (4) Similarly, variable names must be unique within the scope of
   struct and union declarations. Nested struct and union declarations
   create new scopes.

   (5) The discriminant of a union must be of a type that evaluates to
   an integer. That is, "int", "unsigned int", "bool", an enumerated
   type or any typedefed type that evaluates to one of these is legal.
   Also, the case values must be one of the legal values of the
   discriminant.  Finally, a case value may not be specified more than
   once within the scope of a union declaration.

"""
"""

 Being supplemented by RPC structures:

      program-def:
         "program" identifier "{"
            version-def
            version-def *
         "}" "=" constant ";"

      version-def:
         "version" identifier "{"
             procedure-def
             procedure-def *
         "}" "=" constant ";"

      procedure-def:
         proc-return identifier "(" proc-firstarg
           ("," type-specifier )* ")" "=" constant ";"

      proc-return: "void" | type-specifier

      proc-firstarg: "void" | type-specifier


11.3.  Syntax Notes


   o    The following keywords are added and cannot be used as
        identifiers: "program" and "version";

   o    A version name cannot occur more than once within the scope of a
        program definition. Nor can a version number occur more than
        once within the scope of a program definition.

   o    A procedure name cannot occur more than once within the scope of
        a version definition. Nor can a procedure number occur more than
        once within the scope of version definition.

   o    Program identifiers are in the same name space as constant and
        type identifiers.

   o    Only unsigned constants can be assigned to programs, versions
        and procedures.

   o    Current RPC language compilers do not generally support more
        than one type-specifier in procedure argument lists; the usual
        practice is to wrap arguments into a structure.
"""

# Spec above allows the following problems:
# typedef void;
# typedef enum <enum_body> ID[5];  <---Not a problem
# typedef enum <enum_body> ID<>;   <---Not a problem

# Allow to be run straight from package

if __name__ == "__main__":
    if os.path.isfile(os.path.join(sys.path[0], 'lib', 'testmod.py')):
        sys.path.insert(1, os.path.join(sys.path[0], 'lib'))

##########################################################################
#                                                                        #
#                            Lexical analysis                            #
#                                                                        #
##########################################################################

keywords = (
    "bool",
    "case",
    "const",
    "default",
    "double",
    "quadruple",
    "enum",
    "float",
    "hyper",
    "int",
    "opaque",
    "string",
    "struct",
    "switch",
    "typedef",
    "union",
    "unsigned",
    "void",
    # RPC specific
    "program",
    "version",
    # stellar
    "namespace")

# Required by lex.  Each token also allows a function t_<token>.
tokens = tuple([t.upper() for t in keywords]) + (
    "ID",
    "CONST10",
    "CONST8",
    "CONST16",
    # ( ) [ ] { }
    "LPAREN",
    "RPAREN",
    "LBRACKET",
    "RBRACKET",
    "LBRACE",
    "RBRACE",
    # ; : < > * = ,
    "SEMI",
    "COLON",
    "LT",
    "GT",
    "STAR",
    "EQUALS",
    "COMMA")

# t_<name> functions are used by lex.  They are called with t.value==<match
# of rule in comment>, and t.type==<name>.  They expect a return value
# with attribute type=<token>

# Tell lexer to ignore Whitespace
t_ignore = " \t"


def t_ID(t):
    r'[A-Za-z][A-Za-z0-9_]*'
    if t.value in keywords:
        t.type = t.value.upper()
    return t


def t_CONST16(t):
    r'0x[0-9a-fA-F]+'
    return t


def t_CONST8(t):
    r'0[0-7]+'
    return t


def t_CONST10(t):
    r'-?(([1-9]\d*)|0)'
    return t


# Tokens
t_LPAREN = r'\('
t_RPAREN = r'\)'
t_LBRACKET = r'\['
t_RBRACKET = r'\]'
t_LBRACE = r'\{'
t_RBRACE = r'\}'
t_SEMI = r';'
t_COLON = r':'
t_LT = r'<'
t_GT = r'>'
t_STAR = r'\*'
t_EQUALS = r'='
t_COMMA = r','


def t_newline(t):
    r'\n+'
    t.lexer.lineno += t.value.count("\n")


# Comments
def t_comment(t):
    r'/\*(.|\n)*?\*/'
    t.lexer.lineno += t.value.count('\n')


def t_linecomment(t):
    r'%.*\n'
    t.lexer.lineno += 1


def t_stellar_linecomment(t):
    r'//.*\n'
    t.lexer.lineno += 1


def t_error(t):
    print(u"Illegal character {0:s} at line {1:d} type {2:s}".format(
        repr(t.value[0]), t.lexer.lineno, t.type))
    t.lexer.skip(1)


# Build the lexer
lex.lex(debug=0, optimize=False)


##########################################################################
#                                                                        #
#                          Yacc Parsing Info                             #
#                                                                        #
##########################################################################


def p_namespace(t):
    """namespace : NAMESPACE ID LBRACE specification RBRACE"""


def p_specification(t):
    """specification : definition_list"""


def p_definition_list(t):
    """definition_list : definition definition_list
                       | empty"""


def p_definition(t):
    """definition : constant_def
                  | type_def
                  | program_def"""


def p_constant_def(t):
    """constant_def : CONST ID EQUALS constant SEMI"""
    global name_dict
    id = t[2]
    value = t[4]
    lineno = t.lineno(1)
    if id_unique(id, 'constant', lineno):
        name_dict[id] = const_info(id, value, lineno)


def p_constant(t):
    """constant : CONST10
                | CONST8
                | CONST16"""
    if len(t[1]) > 9:
        t[0] = t[1] + 'L'
    else:
        t[0] = t[1]


def p_value(t):
    """value : constant
             | ID"""
    t[0] = t[1]


def p_optional_value(t):
    """optional_value : value
                      | empty"""
    # return value or None.
    t[0] = t[1]
    # Note this must be unsigned
    value = t[0]
    if value is None or value[0].isdigit():
        return
    msg = ''
    if value[0] == '-':
        msg = "Can't use negative index %s" % value
    elif value not in name_dict:
        msg = "Can't derefence index %s" % value
    else:
        data = name_dict[value]
        if data.type != 'const':
            msg = "Can't use non-constant %s %s as index" % (data.type, value)
        elif not data.positive:
            msg = "Can't use negative index %s" % value
    if msg:
        global error_occurred
        error_occurred = True
        print(u"ERROR - {0:s} near line {1:d}".format(msg, t.lineno(1)))


def p_type_def_1(t):
    """type_def : TYPEDEF declaration SEMI"""
    # declarations is a type_info
    d = t[2]
    lineno = t.lineno(1)
    sortno = t.lineno(3) + 0.5
    if d.type == 'void':
        global error_occurred
        error_occurred = True
        print(
            u"ERROR - can't use void in typedef at line {0:d}".format(lineno))
        return
    d.lineno = lineno
    if id_unique(d.id, d.type, lineno):
        if d.type == 'enum':
            info = d.create_enum(lineno, sortno)
        elif d.type == 'struct':
            info = d.create_struct(lineno, sortno)
        elif d.type == 'union':
            info = d.create_union(lineno, sortno)
        else:
            info = d
        name_dict[d.id] = info


def p_type_def_2(t):
    """type_def : ENUM ID enum_body SEMI"""
    id = t[2]
    body = t[3]
    lineno = t.lineno(1)
    sortno = t.lineno(4) + 0.5
    if id_unique(id, 'enum', lineno):
        name_dict[id] = enum_info(id, body, lineno, sortno)


def p_type_def_3(t):
    """type_def : STRUCT ID struct_body SEMI"""
    id = t[2]
    body = t[3]
    lineno = t.lineno(1)
    if id_unique(id, 'struct', lineno):
        name_dict[id] = struct_info(id, body, lineno)


def p_type_def_4(t):
    """type_def : UNION ID union_body SEMI"""
    id = t[2]
    body = t[3]
    lineno = t.lineno(1)
    if id_unique(id, 'union', lineno):
        name_dict[id] = union_info(id, body, lineno)


def p_declaration_1(t):
    """declaration : type_specifier ID"""
    t[1].id = t[2]
    t[0] = t[1]


def p_declaration_2(t):
    """declaration : type_specifier ID LBRACKET value RBRACKET
                   | type_specifier ID LT optional_value GT
                   | OPAQUE ID LBRACKET value RBRACKET
                   | OPAQUE ID LT optional_value GT
                   | STRING ID LT optional_value GT"""
    if not isinstance(t[1], type_info):
        t[1] = type_info(t[1], t.lineno(1))
    t[1].id = t[2]
    t[1].array = True
    if t[3] == '[':
        t[1].fixed = True
    else:
        t[1].fixed = False
    t[1].len = t[4]
    t[0] = t[1]


def p_declaration_3(t):
    """declaration : type_specifier STAR ID"""
    # encode this as the equivalent 'type_specifier ID LT 1 GT'
    if not isinstance(t[1], type_info):
        t[1] = type_info(t[1], t.lineno(1))
    t[1].id = t[3]
    t[1].array = True
    t[1].fixed = False
    t[1].len = '1'
    t[0] = t[1]


def p_declaration_4(t):
    """declaration : VOID"""
    t[0] = type_info(t[1], t.lineno(1))


def p_type_specifier_1(t):
    """type_specifier : UNSIGNED INT
                      | UNSIGNED HYPER"""
    t[0] = type_info('u' + t[2], t.lineno(1))


def p_type_specifier_2(t):
    """type_specifier : INT
                      | HYPER
                      | FLOAT
                      | DOUBLE
                      | QUADRUPLE
                      | BOOL
                      | ID
                      | UNSIGNED
                      | enum_type_spec
                      | struct_type_spec
                      | union_type_spec"""
    # FRED - Note UNSIGNED is not in spec
    if isinstance(t[1], type_info):
        t[0] = t[1]
    else:
        t[0] = type_info(t[1], t.lineno(1))


def p_enum_type_spec(t):
    """enum_type_spec : ENUM enum_body"""
    t[0] = type_info("enum", t.lineno(1), body=t[2])


def p_struct_type_spec(t):
    """struct_type_spec : STRUCT struct_body"""
    t[0] = type_info("struct", t.lineno(1), body=t[2])


def p_union_type_spec(t):
    """union_type_spec : UNION union_body"""
    t[0] = type_info("union", t.lineno(1), body=t[2])


def p_union_body(t):
    """union_body : SWITCH LPAREN declaration RPAREN LBRACE switch_body RBRACE"""
    t[0] = [CaseSpec(['switch'], [t[3]])] + t[6]


def p_switch_body(t):
    """switch_body : case_spec_list default_declaration"""
    # default_declaration is a list of type_info
    t[0] = t[1] + [CaseSpec(['default'], t[2])]


def p_case_spec(t):
    """case_spec : case_statement_list declaration SEMI"""
    # Note a declaration is a type_info
    # case_* are both lists of strings (values)
    t[0] = [CaseSpec(t[1], [t[2]])]


def p_nonempty_lists(t):
    """case_spec_list      : case_spec case_spec_list
                           | case_spec
       case_statement_list : case_statement case_statement_list
                           | case_statement"""
    if len(t) == 2:
        t[0] = t[1]
    else:
        t[0] = t[1] + t[2]


def p_case_statement(t):
    """case_statement : CASE value COLON"""
    t[0] = [t[2]]


def p_default_declaration_1(t):
    """default_declaration : empty"""
    t[0] = []


def p_default_declaration_(t):
    """default_declaration : DEFAULT COLON declaration SEMI"""
    t[0] = [t[3]]


def p_struct_body(t):
    """struct_body : LBRACE declaration_list RBRACE"""
    # Returns a list of type_info declarations
    t[0] = t[2]


def p_declaration_list_1(t):
    """declaration_list : declaration SEMI"""
    t[0] = [t[1]]


def p_declaration_list_2(t):
    """declaration_list : declaration SEMI declaration_list"""
    t[0] = [t[1]] + t[3]


def p_enum_body(t):
    """enum_body : LBRACE enum_constant_list RBRACE"""
    # Returns a list of const_info
    t[0] = t[2]


def p_enum_constant(t):
    """enum_constant : ID EQUALS value"""
    global name_dict, error_occurred
    id = t[1]
    value = t[3]
    lineno = t.lineno(1)
    if id_unique(id, 'enum', lineno):
        info = name_dict[id] = const_info(id, value, lineno, enum=True)
        if not (value[0].isdigit() or value[0] == '-'):
            # We have a name instead of a constant, make sure it is defined
            if value not in name_dict:
                error_occurred = True
                print("ERROR - can't derefence {0:s} at line {1:s}".format(
                    value, lineno))
            elif not isinstance(name_dict[value], const_info):
                error_occurred = True
                print(
                    "ERROR - reference to {0:s} at line {1:s} is not a constant".
                        format(value, lineno))
            else:
                info.positive = name_dict[value].positive
        t[0] = [info]
    else:
        t[0] = []


def p_enum_constant_list_1(t):
    """enum_constant_list : enum_constant"""
    t[0] = t[1]


def p_enum_constant_list_2(t):
    """enum_constant_list : enum_constant COMMA enum_constant_list"""
    t[0] = t[1] + t[3]


def p_empty(t):
    """empty :"""


def p_error(t):
    global error_occurred
    error_occurred = True
    if t:
        print("Syntax error at '{0:s}' (lineno {1:d})".format(
            t.value, t.lineno))
    else:
        print("Syntax error: unexpectedly hit EOF")


#
# RPC specific routines follow
#


def p_program_def(t):
    """program_def : PROGRAM ID LBRACE version_def version_def_list RBRACE EQUALS constant SEMI"""
    print("Ignoring program {0:s} = {1:s}".format(t[2], t[8]))
    global name_dict
    id = t[2]
    value = t[8]
    lineno = t.lineno(1)
    if id_unique(id, 'program', lineno):
        name_dict[id] = const_info(id, value, lineno)


def p_version_def(t):
    """version_def :  VERSION ID LBRACE procedure_def procedure_def_list RBRACE EQUALS constant SEMI"""
    global name_dict
    id = t[2]
    value = t[8]
    lineno = t.lineno(1)
    if id_unique(id, 'version', lineno):
        name_dict[id] = const_info(id, value, lineno)


def p_version_def_list(t):
    """version_def_list : version_def version_def_list
                        | empty"""


def p_procedure_def(t):
    """procedure_def : proc_return ID LPAREN proc_firstarg  type_specifier_list RPAREN EQUALS constant SEMI"""
    global name_dict
    tid = t[2]
    value = t[8]
    lineno = t.lineno(1)
    if id_unique(tid, 'procedure', lineno):
        name_dict[tid] = const_info(tid, value, lineno)


def p_procedure_def_list(t):
    """procedure_def_list : procedure_def procedure_def_list
                          | empty"""


def p_proc_return(t):
    """proc_return : type_specifier
                   | VOID"""


def p_proc_firstarg(t):
    """proc_firstarg : type_specifier
                     | VOID"""


def p_type_specifier_list(t):
    """type_specifier_list : COMMA type_specifier type_specifier_list
                           | empty"""


##########################################################################
#                                                                        #
#                          Global Variables                              #
#                                                                        #
##########################################################################

error_occurred = False  # Parsing of infile status

INDENT = 4  # Number of spaces for each indent level
indent = ' ' * INDENT
indent2 = indent * 2


##########################################################################
#                                                                        #
#                   Helper classes and functions                         #
#                                                                        #
##########################################################################


def id_unique(dict_id, name, lineno):
    """Returns True if dict_id not already used.  Otherwise, invokes error"""
    if dict_id in name_dict:
        global error_occurred
        error_occurred = True
        print(
            "ERROR - {0:s} definition {1:s} at line {2:d} conflicts with {3:s}"
                .format(name, dict_id, lineno, name_dict[dict_id]))
        return False
    else:
        return True


class CaseSpec(object):
    def __init__(self, cases, declarations):
        self.cases = cases
        self.declarations = declarations

    def __str__(self):
        return "cases %s: %s" % (self.cases, self.declarations)


class Info(object):
    def __init__(self):
        self.lineno = None
        self.sortno = None
        self.type = None
        self.array = False
        self.parent = False

    def __str__(self):
        return "%s %s at line %s" % (self.type, self.id, self.lineno)

    def __cmp__(self, other):
        """Sort on lineno, but send None to end"""
        # FRED - not used
        if self.lineno is None and other.lineno is None:
            return 0
        if self.lineno is None:
            return 1
        if other.lineno is None:
            return -1
        if self.lineno < other.lineno:
            return -1
        elif self.lineno == other.lineno:
            return 0
        else:
            return 1

    # def __cmp__(self, other):
    #     """Sort on lineno, but send None to end"""
    #     if self.sortno < other.sortno:
    #         return -1
    #     elif self.sortno == other.sortno:
    #         return 0
    #     else:
    #         return 1

    def const_output(self):
        return None

    def type_output(self):
        return None

    def pack_output(self):
        return None

    def _get_filter(self):
        if use_filters:
            filter1 = "%sif hasattr(self, 'filter_%s'):\n" % (indent2, self.id)
            filter2 = "%sdata = getattr(self, 'filter_%s')(data)\n" % (
                indent * 3, self.id)
            return filter1 + filter2
        else:
            return ''

    def _get_pack_header(self):
        header = "%sdef pack_%s(self, data):\n" % (indent, self.id)
        return header + self._get_filter()

    def unpack_output(self):
        return None

    def _get_unpack_footer(self):
        footer = "%sreturn data\n" % indent2
        return self._get_filter() + footer

    def brackets(self):
        if self.array:
            if self.fixed:
                out = "[%s]"
            else:
                out = "<%s>"
            if self.len is None:
                out = out % ''
            else:
                out = out % self.len
        else:
            out = ''
        return out

    def fullname(self, value):
        """Put 'const.' in front if needed"""
        if value[0].isdigit() or value[0] == '-':
            return value
        else:
            return "const." + value

    def typeinit(self, varlist, prefix=indent):
        initargs = ''.join([", %s=None" % var.id for var in varlist])
        initvars = ''.join([
            "%s%sself.%s = %s\n" % (prefix, indent, var.id, var.id)
            for var in varlist
        ])
        return "%sdef __init__(self%s):\n%s" % (prefix, initargs, initvars)

    #     def typerepr(self, varlist, prefix=indent):
    #         indent2 = prefix + indent
    #         reprbody = ''.join(["%sif self.%s is not None:\n" \
    #                             "%s%sout += ['%s=%%s' %% repr(self.%s)]\n" %
    #                             (indent2, var.id, indent2, indent, var.id, var.id)
    #                             for var in varlist])
    #         return "%sdef __repr__(self):\n%sout = []\n" \
    #                "%s%sreturn '%s(%%s)' %% ', '.join(out)\n" % \
    #                (prefix, indent2, reprbody, indent2, self.id)

    def typerepr(self, varlist, prefix=indent):
        def special(t):
            if t.type in name_dict:
                p = name_dict[t.type]
                if p.parent and p.type == 'enum':
                    if t.array:
                        return "','.join([const.%s.get(x, x) for x in self.%s])" % (
                            p.id, t.id)
                    return "const.%s.get(self.%s, self.%s)" % (p.id, t.id,
                                                               t.id)
            return "repr(self.%s)" % t.id

        indent2 = prefix + indent
        reprbody = ''.join(["%sif self.%s is not None:\n" \
                            "%s%sout += ['%s=%%s' %% %s]\n" %
                            (indent2, var.id, indent2, indent, var.id, special(var))
                            for var in varlist])
        return "%sdef __repr__(self):\n%sout = []\n" \
               "%s%sreturn '%s(%%s)' %% ', '.join(out)\n" \
               "%s__str__ = __repr__\n" % \
               (prefix, indent2, reprbody, indent2, self.id, prefix)

    def _array_pack(self, prefix, data='data'):
        info = globals()["%s_info" % self.type]
        if isinstance(self, info):
            newdata = data
        else:
            # is a type_info (thus not a typedef)
            newdata = "%s.%s" % (data, self.id)
        if self.array:
            newdata = data
            subheader = "%sdef pack_one_%s(self, data):\n" % (prefix, self.id)
            varindent = indent
            if self.fixed:
                array = "%sself.pack_farray(%s, data, pack_one_%s)\n" % \
                        (prefix, self.fullname(self.len), self.id)
            else:
                array = "%sself.pack_array(data, pack_one_%s)\n" % \
                        (prefix, self.id)
                if self.len is not None:
                    limit = "%sif len(data) > %s:\n" \
                            "%s%sraise XDRError('array length too long')\n" % \
                            (prefix, self.fullname(self.len), prefix, varindent)
                    array = limit + array
        else:
            subheader = array = varindent = ''
        return prefix + varindent, newdata, subheader, array

    def _array_unpack(self, prefix, data='data'):
        info = globals()["%s_info" % self.type]
        if isinstance(self, info):
            newdata = data
        else:
            # is a type_info (thus not a typedef)
            newdata = "%s.%s" % (data, self.id)
        if self.array:
            subheader = "%sdef unpack_one_%s(self, data):\n" % \
                        (prefix, self.id)
            varindent = indent
            array = "%s%sreturn data\n" % (prefix, varindent)
            if self.fixed:
                array += "%s%s = self.unpack_farray(%s, unpack_one_%s)\n" % \
                         (prefix, newdata, self.fullname(self.len), self.id)
            else:
                array += "%s%s = self.unpack_array(unpack_one_%s)\n" % \
                         (prefix, newdata, self.id)
                if self.len is not None:
                    limit = "%sif len(%s) > %s:\n" \
                            "%s%sraise XDRError('array length too long')\n" % \
                            (prefix, newdata, self.fullname(self.len), prefix, indent)
                    array += limit
            newdata = 'data'
        else:
            subheader = array = varindent = ''
        return prefix + varindent, newdata, subheader, array

    def packenum(self, prefix, data='data'):
        prefix, data, subheader, array = self._array_pack(prefix, data)
        varlist = ["const.%s" % l.id for l in self.body]
        check = "%sif self.check_enum and %s not in [%s]:\n" \
                "%s%sraise XDRError('value=%%s not in enum %s' %% %s)\n" % \
                (prefix, data, ', '.join(varlist),
                 prefix, indent, self.id, data)
        pack = check + "%sself.pack_int(%s)\n" % (prefix, data)
        return subheader + pack + array

    def unpackenum(self, prefix, data='data'):
        prefix, data, subheader, array = self._array_unpack(prefix, data)
        varlist = ["const.%s" % l.id for l in self.body]
        check = "%sif self.check_enum and %s not in [%s]:\n" \
                "%s%sraise XDRError('value=%%s not in enum %s' %% %s)\n" % \
                (prefix, data, ', '.join(varlist),
                 prefix, indent, self.id, data)
        unpack = "%s%s = self.unpack_int()\n" % (prefix, data)
        return subheader + unpack + check + array

    def packstruct(self, prefix, data='data'):
        prefix, data, subheader, array = self._array_pack(prefix, data)
        pack = ''.join([l.packout(prefix, data) for l in self.body])
        return subheader + pack + array

    def unpackstruct(self, prefix, data='data'):
        prefix, data, subheader, array = self._array_unpack(prefix, data)
        if isinstance(self, struct_info):
            classname = "types.%s" % self.id
        else:
            classname = 'nullclass'
        unpack = "%s%s = %s()\n" % (prefix, data, classname) + \
                 ''.join([l.unpackout(prefix, data) for l in self.body])
        return subheader + unpack + array

    def packunion(self, prefix, data='data'):
        prefix, data, subheader, array = self._array_pack(prefix, data)
        switch = self.body[0].declarations[0]
        pack = switch.packout(prefix, data)
        first = ''
        for l in self.body[1:-1]:
            cases = ' or '.join([
                "%s.%s == %s" % (data, switch.id, self.fullname(c))
                for c in l.cases
            ])
            check = "%s%sif %s:\n" % (prefix, first, cases)
            body = ''.join([d.packout(prefix + indent, data) \
                            for d in l.declarations])
            pack += check + body
            first = 'el'
        default = self.body[-1].declarations
        pack += "%selse:\n" % (prefix)
        if default != []:
            pack += default[0].packout(prefix + indent, data)
        else:
            pack += "%s%sraise XDRError('bad switch=%%s' %% %s.%s)\n" % \
                    (prefix, indent, data, switch.id)
        return subheader + pack + array

    def unpackunion(self, prefix, data='data'):
        prefix, data, subheader, array = self._array_unpack(prefix, data)
        if isinstance(self, union_info):
            classname = "types.%s" % self.id
        else:
            classname = 'nullclass'
        unpack = "%s%s = %s()\n" % (prefix, data, classname)
        switch = self.body[0].declarations[0]
        unpack += switch.unpackout(prefix, data)
        first = ''
        for l in self.body[1:-1]:
            cases = ' or '.join([
                "%s.%s == %s" % (data, switch.id, self.fullname(c))
                for c in l.cases
            ])
            check = "%s%sif %s:\n" % (prefix, first, cases)
            body = ''.join([d.unpackout(prefix + indent, data) \
                            for d in l.declarations])
            ##             if l.declarations[0].id is None:
            ##                  arm = "%s%s%s.switch = None\n" % \
            ##                        (prefix, indent, data)
            ##             else:
            ##                 arm = "%s%s%s.switch = %s.%s\n" % \
            ##                       (prefix, indent, data, data, l.declarations[0].id)
            unpack += check + body  # + arm
            first = 'el'
        default = self.body[-1].declarations
        unpack += "%selse:\n" % (prefix)
        if default != []:
            unpack += ''.join([d.unpackout(prefix + indent, data) \
                               for d in default])
        ##             if default[0].id is None:
        ##                  arm = "%s%s%s.switch = None\n" % \
        ##                        (prefix, indent, data)
        ##             else:
        ##                 arm = "%s%s%s.switch = %s.%s\n" % \
        ##                       (prefix, indent, data, data, default[0].id)
        ##             unpack += arm
        else:
            unpack += "%s%sraise XDRError('bad switch=%%s' %% %s.%s)\n" % \
                      (prefix, indent, data, switch.id)

        return subheader + unpack + array

    def xdrbody(self, prefix=''):
        """Return xdr code for the body (part between braces) of big 3 types"""
        body = ''
        prefix += indent
        if self.type == 'enum':
            body = ''.join(
                ["%s,\n" % l.xdrout(prefix) for l in self.body[:-1]])
            body += "%s\n" % self.body[-1].xdrout(prefix)
        elif self.type == 'struct':
            body = ''.join(["%s\n" % l.xdrout(prefix) for l in self.body])
        elif self.type == 'union':
            for l in self.body[1:-1]:
                body += ''.join(["%scase %s:\n" % (prefix, case) \
                                 for case in l.cases])
                body += ''.join([
                    "%s\n" % d.xdrout(prefix + indent) for d in l.declarations
                ])
            if self.body[-1].declarations:
                body += "%sdefault:\n" % prefix + \
                        ''.join(["%s\n" % d.xdrout(prefix + indent)
                                 for d in self.body[-1].declarations])
        return body


class const_info(Info):
    """The result of 'CONST ID EQUALS constant SEMI' or inside of enum as
    'ID EQUALS value' """

    def __init__(self, id, value, lineno=None, enum=False):
        self.id = id
        self.value = value
        self.positive = value[0] != '-'
        self.lineno = self.sortno = lineno
        self.type = 'const'
        self.enum = enum

    def __repr__(self):
        return "constant %s=%s at line %s" % (self.id, self.value, self.lineno)

    def xdrout(self, prefix=''):
        return "%s%s = %s" % (prefix, self.id, self.value)

    def const_output(self):
        return "%s = %s\n" % (self.id, self.value)


class enum_info(Info):
    """The result of 'TYPEDEF ENUM <enum_body> ID <array> SEMI' or
    'ENUM ID <enum_body> SEMI'

    Note that an enum which is not of above form will appear as a type_info
    """

    def __init__(self, id, body, lineno=None, sortno=None):
        self.id = id
        self.body = body  # list of const_info
        self.lineno = lineno
        if sortno is None:
            self.sortno = self.lineno
        else:
            self.sortno = sortno
        self.type = 'enum'
        self.array = False
        self.parent = True

    def const_output(self):
        body = ''.join(
            ["%s%s : '%s',\n" % (indent, l.value, l.id) for l in self.body])
        return "%s = {\n%s}\n" % (self.id, body)

    def pack_output(self):
        header = self._get_pack_header()
        return header + self.packenum(indent2)

    def unpack_output(self):
        header = "%sdef unpack_%s(self):\n" % (indent, self.id)
        return header + self.unpackenum(indent2) + \
               self._get_unpack_footer()


class struct_info(Info):
    """The result of 'TYPEDEF STRUCT <struct_body> ID <array> SEMI' or
    'STRUCT ID <struct_body> SEMI'

    Note that a struct which is not of above form will appear as a type_info
    """

    def __init__(self, id, body, lineno=None, sortno=None):
        self.id = id
        self.body = body  # list of type_info declarations
        self.lineno = lineno
        if sortno is None:
            self.sortno = self.lineno
        else:
            self.sortno = sortno
        self.type = 'struct'
        self.array = False
        self.parent = True

    def type_output(self):
        comment = '%s# ' % indent
        xdrbody = self.xdrbody(comment)
        xdrdef = "%sXDR definition:\n%sstruct %s {\n%s%s};\n" % \
                 (comment, comment, self.id, xdrbody, comment)
        varlist = [l for l in self.body if l.type != 'void']
        init = self.typeinit(varlist)
        repr = self.typerepr(varlist)
        pass_attr = self.pass_through(varlist)
        to_xdr = self.type_to_xdr()
        from_xdr = self.type_from_xdr()
        return "class %s:\n%s%s\n%s%s\n%s\n%s\n" % \
               (self.id, xdrdef, init, pass_attr, to_xdr, from_xdr, repr)

    def type_to_xdr(self, prefix=indent):
        obj_id = self.id.lower()
        return "%sdef to_xdr(self):\n" \
               "%s%s%s = pack.StellarXDRPacker()\n" \
               "%s%s%s.pack_%s(self)\n" \
               "%s%sreturn base64.b64encode(%s.get_buffer()).decode()\n" % \
               (prefix, prefix, indent, obj_id, prefix, indent, obj_id, self.id, prefix, indent, obj_id)

    def type_from_xdr(self, prefix=indent):
        return "%s@staticmethod\n" \
               "%sdef from_xdr(xdr):\n" \
               "%s%sxdr_decoded = base64.b64decode(xdr)\n" \
               "%s%sxdr_unpacked = pack.StellarXDRUnpacker(xdr_decoded)\n" \
               "%s%sreturn xdr_unpacked.unpack_%s()\n" % \
               (prefix, prefix, prefix, indent, prefix, indent, prefix, indent, self.id)

    def pass_through(self, varlist):
        def check(v):
            if v.array:
                return False
            if v.type == 'struct' or v.type == 'union':
                return True
            if v.type not in name_dict:
                return False
            return check(name_dict[v.type])

        if not allow_attr_passthrough:
            return ''
        candidates = [var for var in varlist if check(var)]
        if len(candidates) == 1:
            return "%sdef __getattr__(self, attr):\n" \
                   "%sif attr == '__setstate__':\n" \
                   "%s%sraise AttributeError\n" \
                   "%sreturn getattr(self.%s, attr)\n\n" % \
                   (indent, indent2, indent2, indent, indent2, candidates[0].id)
        else:
            return ''

    def pack_output(self):
        header = self._get_pack_header()
        return header + self.packstruct(indent2)

    def unpack_output(self):
        header = "%sdef unpack_%s(self):\n" % (indent, self.id)
        return header + self.unpackstruct(indent2) + \
               self._get_unpack_footer()


class union_info(Info):
    """The result of 'TYPEDEF UNION <union_body> ID <array> SEMI' or
    'UNION ID <union_body> SEMI'

    Note that a union which is not of above form will appear as a type_info
    """

    def __init__(self, id, body, lineno=None, sortno=None):
        self.id = id
        self.body = body  # list of type_info declarations
        self.lineno = lineno
        if sortno is None:
            self.sortno = self.lineno
        else:
            self.sortno = sortno
        self.type = 'union'
        self.array = False
        self.parent = True

    def union_getattr(self, prefix=indent):
        # see: https://github.com/StellarCN/py-stellar-base/issues/192
        return "%sdef __getattr__(self, attr):\n" \
               "%s%sif attr == '__setstate__':\n" \
               "%s%s%sraise AttributeError\n" \
               "%s%sreturn getattr(self.switch, attr)\n" % \
               (prefix, prefix, indent, prefix, indent, indent, prefix, indent)

    def union_to_xdr(self, prefix=indent):
        obj_id = self.id.lower()
        return "%sdef to_xdr(self):\n" \
               "%s%s%s = pack.StellarXDRPacker()\n" \
               "%s%s%s.pack_%s(self)\n" \
               "%s%sreturn base64.b64encode(%s.get_buffer()).decode()\n" % \
               (prefix, prefix, indent, obj_id, prefix, indent, obj_id, self.id, prefix, indent, obj_id)

    def union_from_xdr(self, prefix=indent):
        return "%s@staticmethod\n" \
               "%sdef from_xdr(xdr):\n" \
               "%s%sxdr_decoded = base64.b64decode(xdr)\n" \
               "%s%sxdr_unpacked = pack.StellarXDRUnpacker(xdr_decoded)\n" \
               "%s%sreturn xdr_unpacked.unpack_%s()\n" % \
               (prefix, prefix, prefix, indent, prefix, indent, prefix, indent, self.id)

    def union_switch(self, prefix=indent):
        d = '{'
        for l in self.body[1:-1]:
            for c in l.cases:
                id = l.declarations[0].id
                if id is None:
                    lead = ""
                else:
                    lead = "s."
                d += "%s:%s%s," % (self.fullname(c), lead, id)
        d += '}'
        key = self.body[0].declarations[0].id
        default = self.body[-1].declarations
        if default != []:
            id = default[0].id
            if id is None:
                lead = ""
            else:
                lead = "s."
            get = ".get(s.%s, %s%s)" % (key, lead, id)
        else:
            get = "[s.%s]" % key
        return "%sswitch = property(lambda s: %s%s)\n" % (prefix, d, get)

    def type_output(self):
        comment = '%s# ' % indent
        xdrbody = self.xdrbody(comment)
        switch = "%s" % self.body[0].declarations[0].xdrout(comment + ' ' * 13)
        xdrdef = "%sXDR definition:\n%sunion %s switch(%s) {\n%s%s};\n" % \
                 (comment, comment, self.id,
                  switch[13 + len(comment):].rstrip(';'), xdrbody, comment)
        # FRED - need to detect doubles in varlist
        varlist = []
        for c in self.body:
            varlist += [l for l in c.declarations if l.type != 'void']
        init = self.typeinit(varlist)
        repr = self.typerepr(varlist)
        return "class %s:\n%s%s\n%s\n%s\n%s\n%s\n%s\n" % \
               (self.id, xdrdef, init, self.union_switch(), self.union_to_xdr(), self.union_from_xdr(),
                self.union_getattr(), repr)

    def pack_output(self):
        header = self._get_pack_header()
        return header + self.packunion(indent2)

    def unpack_output(self):
        header = "%sdef unpack_%s(self):\n" % (indent, self.id)
        return header + self.unpackunion(indent2) + \
               self._get_unpack_footer()


class type_info(Info):
    def __init__(self, type, lineno=None, body=None):
        self.id = None
        self.type = type
        self.lineno = self.sortno = lineno
        self.body = body
        self.array = False
        if self.array:
            self.len = None
            self.fixed = False
        self.parent = False

    def __str__(self):
        return "%s %s at line %s" % (self.type, self.id, self.lineno)

    def __repr__(self):
        return "%s %s at line %s" % (self.type, self.id, self.lineno)

    def create_enum(self, lineno, sortno=None):
        enum = enum_info(self.id, self.body, lineno, sortno)
        return self._add_array(enum)

    def create_struct(self, lineno, sortno=None):
        struct = struct_info(self.id, self.body, lineno, sortno)
        return self._add_array(struct)

    def create_union(self, lineno, sortno=None):
        union = union_info(self.id, self.body, lineno, sortno)
        return self._add_array(union)

    def _add_array(self, x):
        x.array = self.array
        if self.array:
            x.len = self.len
            x.fixed = self.fixed
        return x

    def xdrout(self, prefix=''):
        if self.type == 'void':
            return "%svoid;" % prefix
        elif self.type == 'enum':
            body = self.xdrbody(prefix)
            name = "%senum {\n%s%s}" % (prefix, body, prefix)

        elif self.type == 'struct':
            body = self.xdrbody(prefix)
            name = "%sstruct {\n%s%s}" % (prefix, body, prefix)

        elif self.type == 'union':
            body = self.xdrbody(prefix)
            switch = "%s" % self.body[0].declarations[0].xdrout(
                prefix + ' ' * 13)
            name = "%sunion switch(%s) {\n%s%s}" % \
                   (prefix, switch[13 + len(prefix):].rstrip(';'), body, prefix)
        else:
            name = prefix + self.type
        return "%s %s%s;" % (name, self.id, self.brackets())

    def packout(self, prefix='', data='data'):
        check = "%sif %s.%s is None:\n" \
                "%s%sraise TypeError('%s.%s == None')\n" % \
                (prefix, data, self.id, prefix, indent, data, self.id)
        if self.type == 'void':
            return prefix + 'pass\n'
        elif self.type == 'struct':
            return check + self.packstruct(prefix, data)

        elif self.type == 'union':
            return check + self.packunion(prefix, data)
        elif self.type == 'enum':
            return check + self.packenum(prefix, data)
        if not self.array:
            return check + \
                   "%sself.pack_%s(%s.%s)\n" % (prefix, self.type, data, self.id)
        return check + self._pack_array(prefix, "%s.%s" % (data, self.id))

    def unpackout(self, prefix='', data='data'):
        if self.type == 'void':
            return prefix + 'pass\n'
        elif self.type == 'struct':
            return self.unpackstruct(prefix, data)

        elif self.type == 'union':
            return self.unpackunion(prefix, data)
        elif self.type == 'enum':
            return self.unpackenum(prefix, data)
        if not self.array:
            return "%s%s.%s = self.unpack_%s()\n" % \
                   (prefix, data, self.id, self.type)
        return self._unpack_array(prefix, "%s.%s" % (data, self.id))

    def type_output(self):
        if not self.array:
            # XXX BUG this needs to be thought out better - probably need to
            # recursively search through casts.  Also, do we need
            # a const_output for the same case with enum?
            cast = name_dict.get(self.type, None)
            if cast:
                if cast.type in ["struct", "union"]:
                    return "%s = %s\n" % (self.id, self.type)
                elif cast.type == "enum":
                    return "%s = const.%s\n" % (self.id, self.type)

    def pack_output(self):
        if not self.array:
            return "%spack_%s = pack_%s\n" % (indent, self.id, self.type)
        header = self._get_pack_header()
        return header + self._pack_array(indent2)

    def unpack_output(self):
        if not self.array:
            return "%sunpack_%s = unpack_%s\n" % (indent, self.id, self.type)
        header = "%sdef unpack_%s(self):\n" % (indent, self.id)
        return header + self._unpack_array(indent2) + \
               self._get_unpack_footer()

    def _pack_array(self, prefix, data='data'):
        if self.fixed or self.len is None:
            limit = ''
        else:
            limit = "%sif len(%s) > %s and self.check_array:\n" \
                    "%s%sraise XDRError(" \
                    "'array length too long for %s')\n" % \
                    (prefix, data, self.fullname(self.len), prefix, indent, data)
        if self.fixed:
            fixchar = 'f'
            fixnum = "%s, " % self.fullname(self.len)
        else:
            fixchar = fixnum = ''
        if self.type == 'string' or self.type == 'opaque':
            type = self.type
            packer = ''
        else:
            type = 'array'
            packer = ", self.pack_%s" % self.type

        pack = "%sself.pack_%s%s(%s%s%s)\n" % \
               (prefix, fixchar, type, fixnum, data, packer)
        return limit + pack

    def _unpack_array(self, prefix, data='data'):
        if self.fixed or self.len is None:
            limit = ''
        else:
            limit = "%sif len(%s) > %s and self.check_array:\n" \
                    "%s%sraise XDRError(" \
                    "'array length too long for %s')\n" % \
                    (prefix, data, self.fullname(self.len), prefix, indent, data)
        if self.fixed:
            fixchar = 'f'
            fixnum = ["%s" % self.fullname(self.len)]
        else:
            fixchar = ''
            fixnum = []
        if self.type == 'string' or self.type == 'opaque':
            type = self.type
            packer = []
        else:
            type = 'array'
            packer = ["self.unpack_%s" % self.type]

        pack = "%s%s = self.unpack_%s%s(%s)\n" % \
               (prefix, data, fixchar, type, ', '.join(fixnum + packer))
        return pack + limit


##########################################################################
#                                                                        #
#                          Main Loop                                     #
#                                                                        #
##########################################################################
name_dict = {}  # list of global names seen, to avoid conflict

use_filters = True  # Option which causes hooks to be generated which
# allows easy subclassing to, for example,
# automatically exand opaque segments
allow_attr_passthrough = True  # Option which allows substructure attrs to
# be referenced directly, in cases where there
# is a unique substructure to search.
pack_header = """\
from . import %s as const
from . import %s as types
import xdrlib
from xdrlib import Error as XDRError

class nullclass(object):
    pass

"""

pack_init = """\
class %sPacker(xdrlib.Packer):
%sdef __init__(self, check_enum=True, check_array=True):
%sxdrlib.Packer.__init__(self)
%sself.check_enum = check_enum
%sself.check_array = check_array

""" % ("%s", indent, indent2, indent2, indent2)

unpack_init = """\
class %sUnpacker(xdrlib.Unpacker):
%sdef __init__(self, data, check_enum=True, check_array=True):
%sxdrlib.Unpacker.__init__(self, data)
%sself.check_enum = check_enum
%sself.check_array = check_array

""" % ("%s", indent, indent2, indent2, indent2)

known_basics = {
    "int": "pack_int",
    # "enum" : "pack_enum",
    "uint": "pack_uint",
    "unsigned": "pack_uint",
    "hyper": "pack_hyper",
    "uhyper": "pack_uhyper",
    "float": "pack_float",
    "double": "pack_double",
    # Note: xdrlib.py does not have a
    # pack_quadruple currently.
    "quadruple": "pack_double",
    "bool": "pack_bool",
    "opaque": "pack_opaque",
    "string": "pack_string"
}
packer_start = ''.join([
    "%spack_%s = xdrlib.Packer.%s\n" % (indent, k, v)
    for k, v in known_basics.items()
])

unpacker_start = ''.join([
    "%sunpack_%s = xdrlib.Unpacker.un%s\n" % (indent, k, v)
    for k, v in known_basics.items()
])


def run(infile, filters=True, pass_attrs=True, debug=False):
    global use_filters, allow_attr_passthrough
    use_filters = filters
    allow_attr_passthrough = pass_attrs

    # Parse the input data with yacc
    global name_dict
    name_dict = {}

    import ply.yacc as yacc
    parser = yacc.yacc(debug=0, write_tables=0)
    name_base = "StellarXDR"

    if os.path.isfile(infile):
        with open(infile) as content:
            data = content.read()
        print("Input file is", infile)
        # name_base = os.path.basename(infile[:infile.rfind(".")])
        name_base = os.path.splitext(os.path.basename(infile))[0]
        parser.parse(data)

    # stellar XDR dir/files
    elif os.path.isdir(infile):
        print("Input dir is", infile)
        # name_base = os.path.basename(os.path.abspath(infile))
        global file_list
        file_list = [
            os.path.abspath(os.path.join(infile, x))
            for x in os.listdir(infile)
        ]
        while file_list:
            file_path = file_list.pop()
            readfile(file_path, parser)

    global constants_file, types_file, packer_file
    # Create output file names (without .py)
    constants_file = name_base + "_const"
    types_file = name_base + "_type"
    packer_file = name_base + "_pack"
    print("Will use output files {0:s}.py, {1:s}.py, and {2:s}.py".format(
        constants_file, types_file, packer_file))

    if error_occurred:
        print("Error occurred, did not write output files")
        return 1

    comment_string = "# Generated by xdrgen.py from {0:s} on {1:s}\n" \
        .format(infile, time.asctime())
    const_fd = open(constants_file + ".py", "w", newline='\n')
    const_fd.write(comment_string)
    type_fd = open(types_file + ".py", "w", newline='\n')
    type_fd.write(comment_string)
    type_fd.write("import base64\n\n")
    type_fd.write("from . import %s as const\n" % constants_file)
    type_fd.write("from . import %s as pack\n" % packer_file)
    pack_fd = open(packer_file + ".py", "w", newline='\n')
    pack_fd.write(comment_string)
    pack_fd.write(pack_header % (constants_file, types_file))
    # pack_fd.write(pack_init % name_base.upper())
    pack_fd.write(pack_init % name_base)
    pack_fd.write(packer_start)

    type_list = sorted(name_dict.values(), key=lambda name: name.sortno)
    # type_list.sort()
    for value in type_list:
        # print value
        output = value.const_output()
        if output is not None:
            # const_fd.write("# **** %s ****\n" % value.id)
            const_fd.write(output)
        output = value.type_output()
        if output is not None:
            # type_fd.write("# **** %s ****\n" % value.id)
            type_fd.write(output)
        output = value.pack_output()
        if output is not None:
            # pack_fd.write("# **** %s %s %s****\n" % (value.id, value.lineno, value.sortno))
            pack_fd.write(output)
            pack_fd.write('\n')
    # pack_fd.write(unpack_init % name_base.upper())
    pack_fd.write(unpack_init % name_base)
    pack_fd.write(unpacker_start)
    for value in type_list:
        output = value.unpack_output()
        if output is not None:
            pack_fd.write(output)
            pack_fd.write('\n')

    const_fd.close()
    type_fd.close()
    pack_fd.close()
    print('Done')
    return


def readfile(file_path, parser):
    global file_list
    with open(file_path) as content:
        data = content.read()
    import re
    s = re.findall('%#.*\/(.*).h\".*\n', data)
    if not s == []:
        file = s[0] + '.x'
        file_depend = os.path.join(os.path.dirname(file_path), file)

        if file_depend in file_list:
            file_list.remove(file_depend)
            readfile(file_depend, parser)
    parser.parse(data)


#
# Section: main
#
if __name__ == "__main__":
    if len(sys.argv) < 2:
        print("Usage: {0:s} <filename | directory>".format(sys.argv[0]))
        sys.exit(1)

    run(sys.argv[1])

# Local variables:
# py-indent-offset: 4
# tab-width: 8
# End:



from decimal import Decimal
from typing import Union

from .operation import Operation
from .utils import check_amount
from ..asset import Asset
from ..xdr import Xdr


class ChangeTrust(Operation):
    """The :class:`ChangeTrust` object, which represents a ChangeTrust
    operation on Stellar's network.

    Creates, updates, or deletes a trustline. For more on trustlines, please
    refer to the `assets documentation
    <https://www.stellar.org/developers/guides/concepts/assets.html>_`.

    Threshold: Medium

    :param asset: The asset for the trust line.
    :param limit: The limit for the asset, defaults to max int64(922337203685.4775807).
        If the limit is set to "0" it deletes the trustline.
    :param source: The source account (defaults to transaction source).

    """

    _DEFAULT_LIMIT = "922337203685.4775807"

    def __init__(
        self, asset: Asset, limit: Union[str, Decimal] = None, source: str = None,
    ) -> None:
        super().__init__(source)
        self.asset: Asset = asset
        if (
            limit is None
        ):  # We don't need this if the user can send the value with correct type.
            self.limit: Union[str, Decimal] = self._DEFAULT_LIMIT
        else:
            check_amount(limit)
            self.limit = limit

    @classmethod
    def type_code(cls) -> int:
        return Xdr.const.CHANGE_TRUST

    def _to_operation_body(self) -> Xdr.nullclass:
        line = self.asset.to_xdr_object()
        limit = Operation.to_xdr_amount(self.limit)

        change_trust_op = Xdr.types.ChangeTrustOp(line, limit)
        body = Xdr.nullclass()
        body.type = Xdr.const.CHANGE_TRUST
        body.changeTrustOp = change_trust_op
        return body

    @classmethod
    def from_xdr_object(
        cls, operation_xdr_object: Xdr.types.Operation
    ) -> "ChangeTrust":
        """Creates a :class:`ChangeTrust` object from an XDR Operation
        object.

        """
        source = Operation.get_source_from_xdr_obj(operation_xdr_object)

        line = Asset.from_xdr_object(operation_xdr_object.body.changeTrustOp.line)
        limit = Operation.from_xdr_amount(operation_xdr_object.body.changeTrustOp.limit)

        op = cls(source=source, asset=line, limit=limit)
        op._source_muxed = Operation.get_source_muxed_from_xdr_obj(operation_xdr_object)
        return op



import base64
import binascii
import struct
from typing import Union

from crc16 import crc16xmodem

from .exceptions import (
    Ed25519SecretSeedInvalidError,
    Ed25519PublicKeyInvalidError,
    ValueError,
    TypeError,
)

__all__ = ["StrKey"]

_version_bytes = {
    "ed25519_public_key": binascii.a2b_hex("30"),  # G 48 6 << 3
    "ed25519_secret_seed": binascii.a2b_hex("90"),  # S 144 18 << 3
    "pre_auth_tx": binascii.a2b_hex("98"),  # T 152 19 << 3
    "sha256_hash": binascii.a2b_hex("b8"),  # X 184 23 << 3
}


class StrKey:
    """StrKey is a helper class that allows encoding and decoding strkey.

    """

    @staticmethod
    def encode_ed25519_public_key(data: bytes) -> str:
        """Encodes data to strkey ed25519 public key.

        :param data: data to encode
        :return: strkey ed25519 public key
        :raises:
            :exc:`ValueError <stellar_sdk.exceptions.ValueError>`
            :exc:`TypeError <stellar_sdk.exceptions.TypeError>`
        """
        return encode_check("ed25519_public_key", data)

    @staticmethod
    def decode_ed25519_public_key(data: str) -> bytes:
        """Decodes strkey ed25519 public key to raw data.

        :param data: strkey ed25519 public key
        :return: raw bytes
        :raises:
            :exc:`Ed25519PublicKeyInvalidError <stellar_sdk.exceptions.Ed25519PublicKeyInvalidError>`
        """
        try:
            return decode_check("ed25519_public_key", data)
        except Exception:
            raise Ed25519PublicKeyInvalidError(
                "Invalid Ed25519 Public Key: {}".format(data)
            )

    @staticmethod
    def is_valid_ed25519_public_key(public_key: str) -> bool:
        """Returns True if the given Stellar public key is a valid ed25519 public key.

        :param public_key: strkey ed25519 public key to check
        :return: `True` if the given Stellar public key is a valid ed25519 public key.
        """
        return is_valid("ed25519_public_key", public_key)

    @staticmethod
    def encode_ed25519_secret_seed(data: bytes) -> str:
        """Encodes data to strkey ed25519 seed.

        :param data: data to encode
        :return: strkey ed25519 seed
        :raises:
            :exc:`ValueError <stellar_sdk.exceptions.ValueError>`
            :exc:`TypeError <stellar_sdk.exceptions.TypeError>`
        """
        return encode_check("ed25519_secret_seed", data)

    @staticmethod
    def decode_ed25519_secret_seed(data: str) -> bytes:
        """Decodes strkey ed25519 seed to raw data.

        :param data: strkey ed25519 seed
        :return: raw bytes
        :raises:
            :exc:`Ed25519SecretSeedInvalidError <stellar_sdk.exceptions.Ed25519SecretSeedInvalidError>`
        """
        try:
            return decode_check("ed25519_secret_seed", data)
        except Exception:
            raise Ed25519SecretSeedInvalidError(f"Invalid Ed25519 Secret Seed: {data}")

    @staticmethod
    def is_valid_ed25519_secret_seed(seed: str):
        """Returns True if the given Stellar secret key is a valid ed25519 secret seed.

        :param seed: strkey ed25519 secret seed
        :return: `True` if the given Stellar secret key is a valid ed25519 secret seed.
        :raises:
            :exc:`ValueError <stellar_sdk.exceptions.ValueError>`
            :exc:`TypeError <stellar_sdk.exceptions.TypeError>`
        """
        return is_valid("ed25519_secret_seed", seed)

    @staticmethod
    def encode_pre_auth_tx(data: bytes) -> str:
        """Encodes data to strkey preAuthTx.

        :param data: data to encode
        :return: strkey preAuthTx
        :raises:
            :exc:`ValueError <stellar_sdk.exceptions.ValueError>`
            :exc:`TypeError <stellar_sdk.exceptions.TypeError>`
        """
        return encode_check("pre_auth_tx", data)

    @staticmethod
    def decode_pre_auth_tx(data: str) -> bytes:
        """Decodes strkey PreAuthTx to raw data.

        :param data: strkey preAuthTx
        :return: raw bytes
        :raises:
            :exc:`ValueError <stellar_sdk.exceptions.ValueError>`
            :exc:`TypeError <stellar_sdk.exceptions.TypeError>`
        """
        return decode_check("pre_auth_tx", data)

    @staticmethod
    def encode_sha256_hash(data: bytes) -> str:
        """Encodes data to strkey sha256 hash.

        :param data: data to encode
        :return: strkey sha256 hash
        :raises:
            :exc:`ValueError <stellar_sdk.exceptions.ValueError>`
            :exc:`TypeError <stellar_sdk.exceptions.TypeError>`
        """
        return encode_check("sha256_hash", data)

    @staticmethod
    def decode_sha256_hash(data: str) -> bytes:
        """Decodes strkey sha256 hash to raw data.

        :param data: strkey sha256 hash
        :return: raw bytes
        :raises:
            :exc:`ValueError <stellar_sdk.exceptions.ValueError>`
            :exc:`TypeError <stellar_sdk.exceptions.TypeError>`
        """
        return decode_check("sha256_hash", data)


def decode_check(version_byte_name: str, encoded: str) -> bytes:
    encoded_data = _bytes_from_decode_data(encoded)
    encoded_data = encoded_data + b"=" * ((4 - len(encoded_data) % 4) % 4)

    try:
        decoded_data = base64.b32decode(encoded_data)
    except binascii.Error:
        raise ValueError("Incorrect padding.")

    if encoded_data != base64.b32encode(decoded_data):  # Is that even possible?
        raise ValueError("Invalid encoded bytes.")

    version_byte = decoded_data[0:1]
    payload = decoded_data[0:-2]
    data = decoded_data[1:-2]
    checksum = decoded_data[-2:]

    expected_version = _version_bytes.get(version_byte_name)
    if expected_version is None:
        raise TypeError(
            f'{version_byte_name} is not a valid version byte name. expected one of "ed25519_public_key", '
            '"ed25519_secret_seed", "pre_auth_tx", "sha256_hash"'
        )

    if version_byte != expected_version:
        raise TypeError(
            f"Invalid version byte. Expected {expected_version}, got {version_byte}"
        )

    expected_checksum = _calculate_checksum(payload)
    if expected_checksum != checksum:
        raise ValueError("Invalid checksum")

    return data


def encode_check(version_byte_name: str, data: bytes) -> str:
    if data is None:
        raise ValueError("cannot encode null data")

    version_byte = _version_bytes.get(version_byte_name)
    if version_byte is None:
        raise TypeError(
            f'{version_byte_name} is not a valid version byte name. expected one of "ed25519_public_key", '
            '"ed25519_secret_seed", "pre_auth_tx", "sha256_hash"'
        )
    payload = version_byte + data
    crc = _calculate_checksum(payload)
    return base64.b32encode(payload + crc).decode("utf-8").rstrip("=")


def is_valid(version_byte_name: str, encoded: str) -> bool:
    if encoded and len(encoded) != 56:
        return False
    try:
        decoded = decode_check(version_byte_name, encoded)
        # if len(decoded) != 32:
        #     return False
    except (ValueError, TypeError):
        return False
    return True


def _bytes_from_decode_data(s: Union[str, bytes, bytearray]) -> bytes:
    """copy from base64._bytes_from_decode_data
    """
    bytes_types = (bytes, bytearray)  # Types acceptable as binary data
    if isinstance(s, str):
        try:
            return s.encode("ascii")
        except UnicodeEncodeError:
            raise ValueError("string argument should contain only ASCII characters")
    if isinstance(s, bytes_types):
        return s
    try:
        return memoryview(s).tobytes()
    except TypeError:
        raise TypeError(
            f"argument should be a bytes-like object or ASCII "
            f"string, not {s.__class__.__name__}"
        ) from None


def _calculate_checksum(payload):
    # This code calculates CRC16-XModem checksum of payload
    checksum = crc16xmodem(payload)
    # Ensure that the checksum is in LSB order.
    checksum = struct.pack("<H", checksum)
    return checksum



import hashlib
import os
from decimal import Decimal, ROUND_FLOOR
from typing import List
from urllib.parse import urlsplit, urlunsplit

from .asset import Asset
from .exceptions import NoApproximationError, TypeError
from .strkey import StrKey
from .xdr import Xdr

MUXED_ACCOUNT_STARTING_LETTER: str = "M"
ED25519_PUBLIC_KEY_STARTING_LETTER: str = "G"


def sha256(data: bytes) -> bytes:
    return hashlib.sha256(data).digest()


def best_rational_approximation(x):
    x = Decimal(x)
    int32_max = Decimal(2147483647)
    fractions = [[Decimal(0), Decimal(1)], [Decimal(1), Decimal(0)]]
    i = 2
    while True:
        if x > int32_max:
            break
        a = x.to_integral_exact(rounding=ROUND_FLOOR)
        f = x - a
        h = a * fractions[i - 1][0] + fractions[i - 2][0]
        k = a * fractions[i - 1][1] + fractions[i - 2][1]
        if h > int32_max or k > int32_max:
            break
        fractions.append([h, k])
        if f.is_zero():
            break
        x = 1 / f
        i = i + 1
    n = fractions[len(fractions) - 1][0]
    d = fractions[len(fractions) - 1][1]
    if n.is_zero() or d.is_zero():
        raise NoApproximationError("Couldn't find approximation.")
    return {"n": int(n), "d": int(d)}


def pack_xdr_array(x):
    if x is None:
        return []
    # if not isinstance(x, list):
    #     return [x]
    return [x]


def unpack_xdr_array(x):
    if not x:
        return None
    return x[0]


def hex_to_bytes(hex_string):
    if isinstance(hex_string, bytes):
        return hex_string
    if isinstance(hex_string, str):
        return bytes.fromhex(hex_string)
    raise TypeError("`hex_string` should be a 32 byte hash or hex encoded string.")


def convert_assets_to_horizon_param(assets: List[Asset]) -> str:
    assets_string = []
    for asset in assets:
        if asset.is_native():
            assets_string.append(asset.type)
        else:
            assets_string.append(f"{asset.code}:{asset.issuer}")
    return ",".join(assets_string)


def urljoin_with_query(base: str, path: str) -> str:
    split_url = urlsplit(base)
    query = split_url.query
    real_path = split_url.path
    if path:
        real_path = os.path.join(split_url.path, path)
    url = urlunsplit(
        (split_url.scheme, split_url.netloc, real_path, query, split_url.fragment)
    )
    return url


def parse_ed25519_account_id_from_muxed_account_xdr_object(
    data: Xdr.types.MuxedAccount,
) -> str:
    if data.ed25519 is not None:
        return StrKey.encode_ed25519_public_key(data.ed25519)
    return StrKey.encode_ed25519_public_key(data.med25519.ed25519)


def is_fee_bump_transaction(xdr: str) -> bool:
    xdr_object = Xdr.types.TransactionEnvelope.from_xdr(xdr)
    te_type = xdr_object.type
    if te_type == Xdr.const.ENVELOPE_TYPE_TX_FEE_BUMP:
        return True
    elif (
        te_type == Xdr.const.ENVELOPE_TYPE_TX
        or te_type == Xdr.const.ENVELOPE_TYPE_TX_V0
    ):
        return False
    else:
        raise ValueError(
            f"This transaction envelope type is not supported, type = {te_type}."
        )



# Generated by xdrgen.py from ../../.xdr/ on Fri Sep 18 20:45:24 2020
KEY_TYPE_ED25519 = 0
KEY_TYPE_PRE_AUTH_TX = 1
KEY_TYPE_HASH_X = 2
KEY_TYPE_MUXED_ED25519 = 0x100
CryptoKeyType = {
    0 : 'KEY_TYPE_ED25519',
    1 : 'KEY_TYPE_PRE_AUTH_TX',
    2 : 'KEY_TYPE_HASH_X',
    0x100 : 'KEY_TYPE_MUXED_ED25519',
}
PUBLIC_KEY_TYPE_ED25519 = KEY_TYPE_ED25519
PublicKeyType = {
    KEY_TYPE_ED25519 : 'PUBLIC_KEY_TYPE_ED25519',
}
SIGNER_KEY_TYPE_ED25519 = KEY_TYPE_ED25519
SIGNER_KEY_TYPE_PRE_AUTH_TX = KEY_TYPE_PRE_AUTH_TX
SIGNER_KEY_TYPE_HASH_X = KEY_TYPE_HASH_X
SignerKeyType = {
    KEY_TYPE_ED25519 : 'SIGNER_KEY_TYPE_ED25519',
    KEY_TYPE_PRE_AUTH_TX : 'SIGNER_KEY_TYPE_PRE_AUTH_TX',
    KEY_TYPE_HASH_X : 'SIGNER_KEY_TYPE_HASH_X',
}
SCP_ST_PREPARE = 0
SCP_ST_CONFIRM = 1
SCP_ST_EXTERNALIZE = 2
SCP_ST_NOMINATE = 3
SCPStatementType = {
    0 : 'SCP_ST_PREPARE',
    1 : 'SCP_ST_CONFIRM',
    2 : 'SCP_ST_EXTERNALIZE',
    3 : 'SCP_ST_NOMINATE',
}
STELLAR_VALUE_BASIC = 0
STELLAR_VALUE_SIGNED = 1
StellarValueType = {
    0 : 'STELLAR_VALUE_BASIC',
    1 : 'STELLAR_VALUE_SIGNED',
}
LEDGER_UPGRADE_VERSION = 1
LEDGER_UPGRADE_BASE_FEE = 2
LEDGER_UPGRADE_MAX_TX_SET_SIZE = 3
LEDGER_UPGRADE_BASE_RESERVE = 4
LedgerUpgradeType = {
    1 : 'LEDGER_UPGRADE_VERSION',
    2 : 'LEDGER_UPGRADE_BASE_FEE',
    3 : 'LEDGER_UPGRADE_MAX_TX_SET_SIZE',
    4 : 'LEDGER_UPGRADE_BASE_RESERVE',
}
METAENTRY = -1
LIVEENTRY = 0
DEADENTRY = 1
INITENTRY = 2
BucketEntryType = {
    -1 : 'METAENTRY',
    0 : 'LIVEENTRY',
    1 : 'DEADENTRY',
    2 : 'INITENTRY',
}
LEDGER_ENTRY_CREATED = 0
LEDGER_ENTRY_UPDATED = 1
LEDGER_ENTRY_REMOVED = 2
LEDGER_ENTRY_STATE = 3
LedgerEntryChangeType = {
    0 : 'LEDGER_ENTRY_CREATED',
    1 : 'LEDGER_ENTRY_UPDATED',
    2 : 'LEDGER_ENTRY_REMOVED',
    3 : 'LEDGER_ENTRY_STATE',
}
ASSET_TYPE_NATIVE = 0
ASSET_TYPE_CREDIT_ALPHANUM4 = 1
ASSET_TYPE_CREDIT_ALPHANUM12 = 2
AssetType = {
    0 : 'ASSET_TYPE_NATIVE',
    1 : 'ASSET_TYPE_CREDIT_ALPHANUM4',
    2 : 'ASSET_TYPE_CREDIT_ALPHANUM12',
}
THRESHOLD_MASTER_WEIGHT = 0
THRESHOLD_LOW = 1
THRESHOLD_MED = 2
THRESHOLD_HIGH = 3
ThresholdIndexes = {
    0 : 'THRESHOLD_MASTER_WEIGHT',
    1 : 'THRESHOLD_LOW',
    2 : 'THRESHOLD_MED',
    3 : 'THRESHOLD_HIGH',
}
ACCOUNT = 0
TRUSTLINE = 1
OFFER = 2
DATA = 3
CLAIMABLE_BALANCE = 4
LedgerEntryType = {
    0 : 'ACCOUNT',
    1 : 'TRUSTLINE',
    2 : 'OFFER',
    3 : 'DATA',
    4 : 'CLAIMABLE_BALANCE',
}
AUTH_REQUIRED_FLAG = 0x1
AUTH_REVOCABLE_FLAG = 0x2
AUTH_IMMUTABLE_FLAG = 0x4
AccountFlags = {
    0x1 : 'AUTH_REQUIRED_FLAG',
    0x2 : 'AUTH_REVOCABLE_FLAG',
    0x4 : 'AUTH_IMMUTABLE_FLAG',
}
MASK_ACCOUNT_FLAGS = 0x7
MAX_SIGNERS = 20
AUTHORIZED_FLAG = 1
AUTHORIZED_TO_MAINTAIN_LIABILITIES_FLAG = 2
TrustLineFlags = {
    1 : 'AUTHORIZED_FLAG',
    2 : 'AUTHORIZED_TO_MAINTAIN_LIABILITIES_FLAG',
}
MASK_TRUSTLINE_FLAGS = 1
MASK_TRUSTLINE_FLAGS_V13 = 3
PASSIVE_FLAG = 1
OfferEntryFlags = {
    1 : 'PASSIVE_FLAG',
}
MASK_OFFERENTRY_FLAGS = 1
CLAIM_PREDICATE_UNCONDITIONAL = 0
CLAIM_PREDICATE_AND = 1
CLAIM_PREDICATE_OR = 2
CLAIM_PREDICATE_NOT = 3
CLAIM_PREDICATE_BEFORE_ABSOLUTE_TIME = 4
CLAIM_PREDICATE_BEFORE_RELATIVE_TIME = 5
ClaimPredicateType = {
    0 : 'CLAIM_PREDICATE_UNCONDITIONAL',
    1 : 'CLAIM_PREDICATE_AND',
    2 : 'CLAIM_PREDICATE_OR',
    3 : 'CLAIM_PREDICATE_NOT',
    4 : 'CLAIM_PREDICATE_BEFORE_ABSOLUTE_TIME',
    5 : 'CLAIM_PREDICATE_BEFORE_RELATIVE_TIME',
}
CLAIMANT_TYPE_V0 = 0
ClaimantType = {
    0 : 'CLAIMANT_TYPE_V0',
}
CLAIMABLE_BALANCE_ID_TYPE_V0 = 0
ClaimableBalanceIDType = {
    0 : 'CLAIMABLE_BALANCE_ID_TYPE_V0',
}
ENVELOPE_TYPE_TX_V0 = 0
ENVELOPE_TYPE_SCP = 1
ENVELOPE_TYPE_TX = 2
ENVELOPE_TYPE_AUTH = 3
ENVELOPE_TYPE_SCPVALUE = 4
ENVELOPE_TYPE_TX_FEE_BUMP = 5
ENVELOPE_TYPE_OP_ID = 6
EnvelopeType = {
    0 : 'ENVELOPE_TYPE_TX_V0',
    1 : 'ENVELOPE_TYPE_SCP',
    2 : 'ENVELOPE_TYPE_TX',
    3 : 'ENVELOPE_TYPE_AUTH',
    4 : 'ENVELOPE_TYPE_SCPVALUE',
    5 : 'ENVELOPE_TYPE_TX_FEE_BUMP',
    6 : 'ENVELOPE_TYPE_OP_ID',
}
CREATE_ACCOUNT = 0
PAYMENT = 1
PATH_PAYMENT_STRICT_RECEIVE = 2
MANAGE_SELL_OFFER = 3
CREATE_PASSIVE_SELL_OFFER = 4
SET_OPTIONS = 5
CHANGE_TRUST = 6
ALLOW_TRUST = 7
ACCOUNT_MERGE = 8
INFLATION = 9
MANAGE_DATA = 10
BUMP_SEQUENCE = 11
MANAGE_BUY_OFFER = 12
PATH_PAYMENT_STRICT_SEND = 13
CREATE_CLAIMABLE_BALANCE = 14
CLAIM_CLAIMABLE_BALANCE = 15
BEGIN_SPONSORING_FUTURE_RESERVES = 16
END_SPONSORING_FUTURE_RESERVES = 17
REVOKE_SPONSORSHIP = 18
OperationType = {
    0 : 'CREATE_ACCOUNT',
    1 : 'PAYMENT',
    2 : 'PATH_PAYMENT_STRICT_RECEIVE',
    3 : 'MANAGE_SELL_OFFER',
    4 : 'CREATE_PASSIVE_SELL_OFFER',
    5 : 'SET_OPTIONS',
    6 : 'CHANGE_TRUST',
    7 : 'ALLOW_TRUST',
    8 : 'ACCOUNT_MERGE',
    9 : 'INFLATION',
    10 : 'MANAGE_DATA',
    11 : 'BUMP_SEQUENCE',
    12 : 'MANAGE_BUY_OFFER',
    13 : 'PATH_PAYMENT_STRICT_SEND',
    14 : 'CREATE_CLAIMABLE_BALANCE',
    15 : 'CLAIM_CLAIMABLE_BALANCE',
    16 : 'BEGIN_SPONSORING_FUTURE_RESERVES',
    17 : 'END_SPONSORING_FUTURE_RESERVES',
    18 : 'REVOKE_SPONSORSHIP',
}
REVOKE_SPONSORSHIP_LEDGER_ENTRY = 0
REVOKE_SPONSORSHIP_SIGNER = 1
RevokeSponsorshipType = {
    0 : 'REVOKE_SPONSORSHIP_LEDGER_ENTRY',
    1 : 'REVOKE_SPONSORSHIP_SIGNER',
}
MEMO_NONE = 0
MEMO_TEXT = 1
MEMO_ID = 2
MEMO_HASH = 3
MEMO_RETURN = 4
MemoType = {
    0 : 'MEMO_NONE',
    1 : 'MEMO_TEXT',
    2 : 'MEMO_ID',
    3 : 'MEMO_HASH',
    4 : 'MEMO_RETURN',
}
MAX_OPS_PER_TX = 100
CREATE_ACCOUNT_SUCCESS = 0
CREATE_ACCOUNT_MALFORMED = -1
CREATE_ACCOUNT_UNDERFUNDED = -2
CREATE_ACCOUNT_LOW_RESERVE = -3
CREATE_ACCOUNT_ALREADY_EXIST = -4
CreateAccountResultCode = {
    0 : 'CREATE_ACCOUNT_SUCCESS',
    -1 : 'CREATE_ACCOUNT_MALFORMED',
    -2 : 'CREATE_ACCOUNT_UNDERFUNDED',
    -3 : 'CREATE_ACCOUNT_LOW_RESERVE',
    -4 : 'CREATE_ACCOUNT_ALREADY_EXIST',
}
PAYMENT_SUCCESS = 0
PAYMENT_MALFORMED = -1
PAYMENT_UNDERFUNDED = -2
PAYMENT_SRC_NO_TRUST = -3
PAYMENT_SRC_NOT_AUTHORIZED = -4
PAYMENT_NO_DESTINATION = -5
PAYMENT_NO_TRUST = -6
PAYMENT_NOT_AUTHORIZED = -7
PAYMENT_LINE_FULL = -8
PAYMENT_NO_ISSUER = -9
PaymentResultCode = {
    0 : 'PAYMENT_SUCCESS',
    -1 : 'PAYMENT_MALFORMED',
    -2 : 'PAYMENT_UNDERFUNDED',
    -3 : 'PAYMENT_SRC_NO_TRUST',
    -4 : 'PAYMENT_SRC_NOT_AUTHORIZED',
    -5 : 'PAYMENT_NO_DESTINATION',
    -6 : 'PAYMENT_NO_TRUST',
    -7 : 'PAYMENT_NOT_AUTHORIZED',
    -8 : 'PAYMENT_LINE_FULL',
    -9 : 'PAYMENT_NO_ISSUER',
}
PATH_PAYMENT_STRICT_RECEIVE_SUCCESS = 0
PATH_PAYMENT_STRICT_RECEIVE_MALFORMED = -1
PATH_PAYMENT_STRICT_RECEIVE_UNDERFUNDED = -2
PATH_PAYMENT_STRICT_RECEIVE_SRC_NO_TRUST = -3
PATH_PAYMENT_STRICT_RECEIVE_SRC_NOT_AUTHORIZED = -4
PATH_PAYMENT_STRICT_RECEIVE_NO_DESTINATION = -5
PATH_PAYMENT_STRICT_RECEIVE_NO_TRUST = -6
PATH_PAYMENT_STRICT_RECEIVE_NOT_AUTHORIZED = -7
PATH_PAYMENT_STRICT_RECEIVE_LINE_FULL = -8
PATH_PAYMENT_STRICT_RECEIVE_NO_ISSUER = -9
PATH_PAYMENT_STRICT_RECEIVE_TOO_FEW_OFFERS = -10
PATH_PAYMENT_STRICT_RECEIVE_OFFER_CROSS_SELF = -11
PATH_PAYMENT_STRICT_RECEIVE_OVER_SENDMAX = -12
PathPaymentStrictReceiveResultCode = {
    0 : 'PATH_PAYMENT_STRICT_RECEIVE_SUCCESS',
    -1 : 'PATH_PAYMENT_STRICT_RECEIVE_MALFORMED',
    -2 : 'PATH_PAYMENT_STRICT_RECEIVE_UNDERFUNDED',
    -3 : 'PATH_PAYMENT_STRICT_RECEIVE_SRC_NO_TRUST',
    -4 : 'PATH_PAYMENT_STRICT_RECEIVE_SRC_NOT_AUTHORIZED',
    -5 : 'PATH_PAYMENT_STRICT_RECEIVE_NO_DESTINATION',
    -6 : 'PATH_PAYMENT_STRICT_RECEIVE_NO_TRUST',
    -7 : 'PATH_PAYMENT_STRICT_RECEIVE_NOT_AUTHORIZED',
    -8 : 'PATH_PAYMENT_STRICT_RECEIVE_LINE_FULL',
    -9 : 'PATH_PAYMENT_STRICT_RECEIVE_NO_ISSUER',
    -10 : 'PATH_PAYMENT_STRICT_RECEIVE_TOO_FEW_OFFERS',
    -11 : 'PATH_PAYMENT_STRICT_RECEIVE_OFFER_CROSS_SELF',
    -12 : 'PATH_PAYMENT_STRICT_RECEIVE_OVER_SENDMAX',
}
PATH_PAYMENT_STRICT_SEND_SUCCESS = 0
PATH_PAYMENT_STRICT_SEND_MALFORMED = -1
PATH_PAYMENT_STRICT_SEND_UNDERFUNDED = -2
PATH_PAYMENT_STRICT_SEND_SRC_NO_TRUST = -3
PATH_PAYMENT_STRICT_SEND_SRC_NOT_AUTHORIZED = -4
PATH_PAYMENT_STRICT_SEND_NO_DESTINATION = -5
PATH_PAYMENT_STRICT_SEND_NO_TRUST = -6
PATH_PAYMENT_STRICT_SEND_NOT_AUTHORIZED = -7
PATH_PAYMENT_STRICT_SEND_LINE_FULL = -8
PATH_PAYMENT_STRICT_SEND_NO_ISSUER = -9
PATH_PAYMENT_STRICT_SEND_TOO_FEW_OFFERS = -10
PATH_PAYMENT_STRICT_SEND_OFFER_CROSS_SELF = -11
PATH_PAYMENT_STRICT_SEND_UNDER_DESTMIN = -12
PathPaymentStrictSendResultCode = {
    0 : 'PATH_PAYMENT_STRICT_SEND_SUCCESS',
    -1 : 'PATH_PAYMENT_STRICT_SEND_MALFORMED',
    -2 : 'PATH_PAYMENT_STRICT_SEND_UNDERFUNDED',
    -3 : 'PATH_PAYMENT_STRICT_SEND_SRC_NO_TRUST',
    -4 : 'PATH_PAYMENT_STRICT_SEND_SRC_NOT_AUTHORIZED',
    -5 : 'PATH_PAYMENT_STRICT_SEND_NO_DESTINATION',
    -6 : 'PATH_PAYMENT_STRICT_SEND_NO_TRUST',
    -7 : 'PATH_PAYMENT_STRICT_SEND_NOT_AUTHORIZED',
    -8 : 'PATH_PAYMENT_STRICT_SEND_LINE_FULL',
    -9 : 'PATH_PAYMENT_STRICT_SEND_NO_ISSUER',
    -10 : 'PATH_PAYMENT_STRICT_SEND_TOO_FEW_OFFERS',
    -11 : 'PATH_PAYMENT_STRICT_SEND_OFFER_CROSS_SELF',
    -12 : 'PATH_PAYMENT_STRICT_SEND_UNDER_DESTMIN',
}
MANAGE_SELL_OFFER_SUCCESS = 0
MANAGE_SELL_OFFER_MALFORMED = -1
MANAGE_SELL_OFFER_SELL_NO_TRUST = -2
MANAGE_SELL_OFFER_BUY_NO_TRUST = -3
MANAGE_SELL_OFFER_SELL_NOT_AUTHORIZED = -4
MANAGE_SELL_OFFER_BUY_NOT_AUTHORIZED = -5
MANAGE_SELL_OFFER_LINE_FULL = -6
MANAGE_SELL_OFFER_UNDERFUNDED = -7
MANAGE_SELL_OFFER_CROSS_SELF = -8
MANAGE_SELL_OFFER_SELL_NO_ISSUER = -9
MANAGE_SELL_OFFER_BUY_NO_ISSUER = -10
MANAGE_SELL_OFFER_NOT_FOUND = -11
MANAGE_SELL_OFFER_LOW_RESERVE = -12
ManageSellOfferResultCode = {
    0 : 'MANAGE_SELL_OFFER_SUCCESS',
    -1 : 'MANAGE_SELL_OFFER_MALFORMED',
    -2 : 'MANAGE_SELL_OFFER_SELL_NO_TRUST',
    -3 : 'MANAGE_SELL_OFFER_BUY_NO_TRUST',
    -4 : 'MANAGE_SELL_OFFER_SELL_NOT_AUTHORIZED',
    -5 : 'MANAGE_SELL_OFFER_BUY_NOT_AUTHORIZED',
    -6 : 'MANAGE_SELL_OFFER_LINE_FULL',
    -7 : 'MANAGE_SELL_OFFER_UNDERFUNDED',
    -8 : 'MANAGE_SELL_OFFER_CROSS_SELF',
    -9 : 'MANAGE_SELL_OFFER_SELL_NO_ISSUER',
    -10 : 'MANAGE_SELL_OFFER_BUY_NO_ISSUER',
    -11 : 'MANAGE_SELL_OFFER_NOT_FOUND',
    -12 : 'MANAGE_SELL_OFFER_LOW_RESERVE',
}
MANAGE_OFFER_CREATED = 0
MANAGE_OFFER_UPDATED = 1
MANAGE_OFFER_DELETED = 2
ManageOfferEffect = {
    0 : 'MANAGE_OFFER_CREATED',
    1 : 'MANAGE_OFFER_UPDATED',
    2 : 'MANAGE_OFFER_DELETED',
}
MANAGE_BUY_OFFER_SUCCESS = 0
MANAGE_BUY_OFFER_MALFORMED = -1
MANAGE_BUY_OFFER_SELL_NO_TRUST = -2
MANAGE_BUY_OFFER_BUY_NO_TRUST = -3
MANAGE_BUY_OFFER_SELL_NOT_AUTHORIZED = -4
MANAGE_BUY_OFFER_BUY_NOT_AUTHORIZED = -5
MANAGE_BUY_OFFER_LINE_FULL = -6
MANAGE_BUY_OFFER_UNDERFUNDED = -7
MANAGE_BUY_OFFER_CROSS_SELF = -8
MANAGE_BUY_OFFER_SELL_NO_ISSUER = -9
MANAGE_BUY_OFFER_BUY_NO_ISSUER = -10
MANAGE_BUY_OFFER_NOT_FOUND = -11
MANAGE_BUY_OFFER_LOW_RESERVE = -12
ManageBuyOfferResultCode = {
    0 : 'MANAGE_BUY_OFFER_SUCCESS',
    -1 : 'MANAGE_BUY_OFFER_MALFORMED',
    -2 : 'MANAGE_BUY_OFFER_SELL_NO_TRUST',
    -3 : 'MANAGE_BUY_OFFER_BUY_NO_TRUST',
    -4 : 'MANAGE_BUY_OFFER_SELL_NOT_AUTHORIZED',
    -5 : 'MANAGE_BUY_OFFER_BUY_NOT_AUTHORIZED',
    -6 : 'MANAGE_BUY_OFFER_LINE_FULL',
    -7 : 'MANAGE_BUY_OFFER_UNDERFUNDED',
    -8 : 'MANAGE_BUY_OFFER_CROSS_SELF',
    -9 : 'MANAGE_BUY_OFFER_SELL_NO_ISSUER',
    -10 : 'MANAGE_BUY_OFFER_BUY_NO_ISSUER',
    -11 : 'MANAGE_BUY_OFFER_NOT_FOUND',
    -12 : 'MANAGE_BUY_OFFER_LOW_RESERVE',
}
SET_OPTIONS_SUCCESS = 0
SET_OPTIONS_LOW_RESERVE = -1
SET_OPTIONS_TOO_MANY_SIGNERS = -2
SET_OPTIONS_BAD_FLAGS = -3
SET_OPTIONS_INVALID_INFLATION = -4
SET_OPTIONS_CANT_CHANGE = -5
SET_OPTIONS_UNKNOWN_FLAG = -6
SET_OPTIONS_THRESHOLD_OUT_OF_RANGE = -7
SET_OPTIONS_BAD_SIGNER = -8
SET_OPTIONS_INVALID_HOME_DOMAIN = -9
SetOptionsResultCode = {
    0 : 'SET_OPTIONS_SUCCESS',
    -1 : 'SET_OPTIONS_LOW_RESERVE',
    -2 : 'SET_OPTIONS_TOO_MANY_SIGNERS',
    -3 : 'SET_OPTIONS_BAD_FLAGS',
    -4 : 'SET_OPTIONS_INVALID_INFLATION',
    -5 : 'SET_OPTIONS_CANT_CHANGE',
    -6 : 'SET_OPTIONS_UNKNOWN_FLAG',
    -7 : 'SET_OPTIONS_THRESHOLD_OUT_OF_RANGE',
    -8 : 'SET_OPTIONS_BAD_SIGNER',
    -9 : 'SET_OPTIONS_INVALID_HOME_DOMAIN',
}
CHANGE_TRUST_SUCCESS = 0
CHANGE_TRUST_MALFORMED = -1
CHANGE_TRUST_NO_ISSUER = -2
CHANGE_TRUST_INVALID_LIMIT = -3
CHANGE_TRUST_LOW_RESERVE = -4
CHANGE_TRUST_SELF_NOT_ALLOWED = -5
ChangeTrustResultCode = {
    0 : 'CHANGE_TRUST_SUCCESS',
    -1 : 'CHANGE_TRUST_MALFORMED',
    -2 : 'CHANGE_TRUST_NO_ISSUER',
    -3 : 'CHANGE_TRUST_INVALID_LIMIT',
    -4 : 'CHANGE_TRUST_LOW_RESERVE',
    -5 : 'CHANGE_TRUST_SELF_NOT_ALLOWED',
}
ALLOW_TRUST_SUCCESS = 0
ALLOW_TRUST_MALFORMED = -1
ALLOW_TRUST_NO_TRUST_LINE = -2
ALLOW_TRUST_TRUST_NOT_REQUIRED = -3
ALLOW_TRUST_CANT_REVOKE = -4
ALLOW_TRUST_SELF_NOT_ALLOWED = -5
AllowTrustResultCode = {
    0 : 'ALLOW_TRUST_SUCCESS',
    -1 : 'ALLOW_TRUST_MALFORMED',
    -2 : 'ALLOW_TRUST_NO_TRUST_LINE',
    -3 : 'ALLOW_TRUST_TRUST_NOT_REQUIRED',
    -4 : 'ALLOW_TRUST_CANT_REVOKE',
    -5 : 'ALLOW_TRUST_SELF_NOT_ALLOWED',
}
ACCOUNT_MERGE_SUCCESS = 0
ACCOUNT_MERGE_MALFORMED = -1
ACCOUNT_MERGE_NO_ACCOUNT = -2
ACCOUNT_MERGE_IMMUTABLE_SET = -3
ACCOUNT_MERGE_HAS_SUB_ENTRIES = -4
ACCOUNT_MERGE_SEQNUM_TOO_FAR = -5
ACCOUNT_MERGE_DEST_FULL = -6
ACCOUNT_MERGE_IS_SPONSOR = -7
AccountMergeResultCode = {
    0 : 'ACCOUNT_MERGE_SUCCESS',
    -1 : 'ACCOUNT_MERGE_MALFORMED',
    -2 : 'ACCOUNT_MERGE_NO_ACCOUNT',
    -3 : 'ACCOUNT_MERGE_IMMUTABLE_SET',
    -4 : 'ACCOUNT_MERGE_HAS_SUB_ENTRIES',
    -5 : 'ACCOUNT_MERGE_SEQNUM_TOO_FAR',
    -6 : 'ACCOUNT_MERGE_DEST_FULL',
    -7 : 'ACCOUNT_MERGE_IS_SPONSOR',
}
INFLATION_SUCCESS = 0
INFLATION_NOT_TIME = -1
InflationResultCode = {
    0 : 'INFLATION_SUCCESS',
    -1 : 'INFLATION_NOT_TIME',
}
MANAGE_DATA_SUCCESS = 0
MANAGE_DATA_NOT_SUPPORTED_YET = -1
MANAGE_DATA_NAME_NOT_FOUND = -2
MANAGE_DATA_LOW_RESERVE = -3
MANAGE_DATA_INVALID_NAME = -4
ManageDataResultCode = {
    0 : 'MANAGE_DATA_SUCCESS',
    -1 : 'MANAGE_DATA_NOT_SUPPORTED_YET',
    -2 : 'MANAGE_DATA_NAME_NOT_FOUND',
    -3 : 'MANAGE_DATA_LOW_RESERVE',
    -4 : 'MANAGE_DATA_INVALID_NAME',
}
BUMP_SEQUENCE_SUCCESS = 0
BUMP_SEQUENCE_BAD_SEQ = -1
BumpSequenceResultCode = {
    0 : 'BUMP_SEQUENCE_SUCCESS',
    -1 : 'BUMP_SEQUENCE_BAD_SEQ',
}
CREATE_CLAIMABLE_BALANCE_SUCCESS = 0
CREATE_CLAIMABLE_BALANCE_MALFORMED = -1
CREATE_CLAIMABLE_BALANCE_LOW_RESERVE = -2
CREATE_CLAIMABLE_BALANCE_NO_TRUST = -3
CREATE_CLAIMABLE_BALANCE_NOT_AUTHORIZED = -4
CREATE_CLAIMABLE_BALANCE_UNDERFUNDED = -5
CreateClaimableBalanceResultCode = {
    0 : 'CREATE_CLAIMABLE_BALANCE_SUCCESS',
    -1 : 'CREATE_CLAIMABLE_BALANCE_MALFORMED',
    -2 : 'CREATE_CLAIMABLE_BALANCE_LOW_RESERVE',
    -3 : 'CREATE_CLAIMABLE_BALANCE_NO_TRUST',
    -4 : 'CREATE_CLAIMABLE_BALANCE_NOT_AUTHORIZED',
    -5 : 'CREATE_CLAIMABLE_BALANCE_UNDERFUNDED',
}
CLAIM_CLAIMABLE_BALANCE_SUCCESS = 0
CLAIM_CLAIMABLE_BALANCE_DOES_NOT_EXIST = -1
CLAIM_CLAIMABLE_BALANCE_CANNOT_CLAIM = -2
CLAIM_CLAIMABLE_BALANCE_LINE_FULL = -3
CLAIM_CLAIMABLE_BALANCE_NO_TRUST = -4
CLAIM_CLAIMABLE_BALANCE_NOT_AUTHORIZED = -5
ClaimClaimableBalanceResultCode = {
    0 : 'CLAIM_CLAIMABLE_BALANCE_SUCCESS',
    -1 : 'CLAIM_CLAIMABLE_BALANCE_DOES_NOT_EXIST',
    -2 : 'CLAIM_CLAIMABLE_BALANCE_CANNOT_CLAIM',
    -3 : 'CLAIM_CLAIMABLE_BALANCE_LINE_FULL',
    -4 : 'CLAIM_CLAIMABLE_BALANCE_NO_TRUST',
    -5 : 'CLAIM_CLAIMABLE_BALANCE_NOT_AUTHORIZED',
}
BEGIN_SPONSORING_FUTURE_RESERVES_SUCCESS = 0
BEGIN_SPONSORING_FUTURE_RESERVES_MALFORMED = -1
BEGIN_SPONSORING_FUTURE_RESERVES_ALREADY_SPONSORED = -2
BEGIN_SPONSORING_FUTURE_RESERVES_RECURSIVE = -3
BeginSponsoringFutureReservesResultCode = {
    0 : 'BEGIN_SPONSORING_FUTURE_RESERVES_SUCCESS',
    -1 : 'BEGIN_SPONSORING_FUTURE_RESERVES_MALFORMED',
    -2 : 'BEGIN_SPONSORING_FUTURE_RESERVES_ALREADY_SPONSORED',
    -3 : 'BEGIN_SPONSORING_FUTURE_RESERVES_RECURSIVE',
}
END_SPONSORING_FUTURE_RESERVES_SUCCESS = 0
END_SPONSORING_FUTURE_RESERVES_NOT_SPONSORED = -1
EndSponsoringFutureReservesResultCode = {
    0 : 'END_SPONSORING_FUTURE_RESERVES_SUCCESS',
    -1 : 'END_SPONSORING_FUTURE_RESERVES_NOT_SPONSORED',
}
REVOKE_SPONSORSHIP_SUCCESS = 0
REVOKE_SPONSORSHIP_DOES_NOT_EXIST = -1
REVOKE_SPONSORSHIP_NOT_SPONSOR = -2
REVOKE_SPONSORSHIP_LOW_RESERVE = -3
REVOKE_SPONSORSHIP_ONLY_TRANSFERABLE = -4
RevokeSponsorshipResultCode = {
    0 : 'REVOKE_SPONSORSHIP_SUCCESS',
    -1 : 'REVOKE_SPONSORSHIP_DOES_NOT_EXIST',
    -2 : 'REVOKE_SPONSORSHIP_NOT_SPONSOR',
    -3 : 'REVOKE_SPONSORSHIP_LOW_RESERVE',
    -4 : 'REVOKE_SPONSORSHIP_ONLY_TRANSFERABLE',
}
opINNER = 0
opBAD_AUTH = -1
opNO_ACCOUNT = -2
opNOT_SUPPORTED = -3
opTOO_MANY_SUBENTRIES = -4
opEXCEEDED_WORK_LIMIT = -5
opTOO_MANY_SPONSORING = -6
OperationResultCode = {
    0 : 'opINNER',
    -1 : 'opBAD_AUTH',
    -2 : 'opNO_ACCOUNT',
    -3 : 'opNOT_SUPPORTED',
    -4 : 'opTOO_MANY_SUBENTRIES',
    -5 : 'opEXCEEDED_WORK_LIMIT',
    -6 : 'opTOO_MANY_SPONSORING',
}
txFEE_BUMP_INNER_SUCCESS = 1
txSUCCESS = 0
txFAILED = -1
txTOO_EARLY = -2
txTOO_LATE = -3
txMISSING_OPERATION = -4
txBAD_SEQ = -5
txBAD_AUTH = -6
txINSUFFICIENT_BALANCE = -7
txNO_ACCOUNT = -8
txINSUFFICIENT_FEE = -9
txBAD_AUTH_EXTRA = -10
txINTERNAL_ERROR = -11
txNOT_SUPPORTED = -12
txFEE_BUMP_INNER_FAILED = -13
txBAD_SPONSORSHIP = -14
TransactionResultCode = {
    1 : 'txFEE_BUMP_INNER_SUCCESS',
    0 : 'txSUCCESS',
    -1 : 'txFAILED',
    -2 : 'txTOO_EARLY',
    -3 : 'txTOO_LATE',
    -4 : 'txMISSING_OPERATION',
    -5 : 'txBAD_SEQ',
    -6 : 'txBAD_AUTH',
    -7 : 'txINSUFFICIENT_BALANCE',
    -8 : 'txNO_ACCOUNT',
    -9 : 'txINSUFFICIENT_FEE',
    -10 : 'txBAD_AUTH_EXTRA',
    -11 : 'txINTERNAL_ERROR',
    -12 : 'txNOT_SUPPORTED',
    -13 : 'txFEE_BUMP_INNER_FAILED',
    -14 : 'txBAD_SPONSORSHIP',
}
ERR_MISC = 0
ERR_DATA = 1
ERR_CONF = 2
ERR_AUTH = 3
ERR_LOAD = 4
ErrorCode = {
    0 : 'ERR_MISC',
    1 : 'ERR_DATA',
    2 : 'ERR_CONF',
    3 : 'ERR_AUTH',
    4 : 'ERR_LOAD',
}
IPv4 = 0
IPv6 = 1
IPAddrType = {
    0 : 'IPv4',
    1 : 'IPv6',
}
ERROR_MSG = 0
AUTH = 2
DONT_HAVE = 3
GET_PEERS = 4
PEERS = 5
GET_TX_SET = 6
TX_SET = 7
TRANSACTION = 8
GET_SCP_QUORUMSET = 9
SCP_QUORUMSET = 10
SCP_MESSAGE = 11
GET_SCP_STATE = 12
HELLO = 13
SURVEY_REQUEST = 14
SURVEY_RESPONSE = 15
MessageType = {
    0 : 'ERROR_MSG',
    2 : 'AUTH',
    3 : 'DONT_HAVE',
    4 : 'GET_PEERS',
    5 : 'PEERS',
    6 : 'GET_TX_SET',
    7 : 'TX_SET',
    8 : 'TRANSACTION',
    9 : 'GET_SCP_QUORUMSET',
    10 : 'SCP_QUORUMSET',
    11 : 'SCP_MESSAGE',
    12 : 'GET_SCP_STATE',
    13 : 'HELLO',
    14 : 'SURVEY_REQUEST',
    15 : 'SURVEY_RESPONSE',
}
SURVEY_TOPOLOGY = 0
SurveyMessageCommandType = {
    0 : 'SURVEY_TOPOLOGY',
}



from decimal import Decimal
from typing import Union

from .operation import Operation
from .utils import check_price, check_amount
from ..asset import Asset
from ..price import Price
from ..xdr import Xdr


class ManageBuyOffer(Operation):
    """The :class:`ManageBuyOffer` object, which represents a ManageBuyOffer
    operation on Stellar's network.

    Creates, updates, or deletes an buy offer.

    If you want to create a new offer set Offer ID to 0.

    If you want to update an existing offer set Offer ID to existing offer ID.

    If you want to delete an existing offer set Offer ID to existing offer ID
    and set Amount to 0.

    Threshold: Medium

    :param selling: What you're selling.
    :param buying: What you're buying.
    :param amount: Amount being bought. if set to 0, delete the offer.
    :param price: Price of thing being bought in terms of what you are selling.
    :param offer_id: If `0`, will create a new offer (default). Otherwise,
        edits an existing offer.
    :param source: The source account (defaults to transaction source).

    """

    def __init__(
        self,
        selling: Asset,
        buying: Asset,
        amount: Union[str, Decimal],
        price: Union[Price, str, Decimal],
        offer_id: int = 0,
        source: str = None,
    ) -> None:
        super().__init__(source)
        check_amount(amount)
        check_price(price)
        self.selling: Asset = selling
        self.buying: Asset = buying
        self.amount: Union[str, Decimal] = amount
        self.price: Union[Price, str, Decimal] = price
        self.offer_id: int = offer_id

    @classmethod
    def type_code(cls) -> int:
        return Xdr.const.MANAGE_BUY_OFFER

    def _to_operation_body(self) -> Xdr.nullclass:
        selling = self.selling.to_xdr_object()
        buying = self.buying.to_xdr_object()

        if isinstance(self.price, Price):
            price_fraction = self.price
        else:
            price_fraction = Price.from_raw_price(self.price)

        price = price_fraction.to_xdr_object()

        amount = Operation.to_xdr_amount(self.amount)

        manage_buy_offer_op = Xdr.types.ManageBuyOfferOp(
            selling, buying, amount, price, self.offer_id
        )
        body = Xdr.nullclass()
        body.type = Xdr.const.MANAGE_BUY_OFFER
        body.manageBuyOfferOp = manage_buy_offer_op
        return body

    @classmethod
    def from_xdr_object(
        cls, operation_xdr_object: Xdr.types.Operation
    ) -> "ManageBuyOffer":
        """Creates a :class:`ManageBuyOffer` object from an XDR Operation object.

        """
        source = Operation.get_source_from_xdr_obj(operation_xdr_object)

        selling = Asset.from_xdr_object(
            operation_xdr_object.body.manageBuyOfferOp.selling
        )
        buying = Asset.from_xdr_object(
            operation_xdr_object.body.manageBuyOfferOp.buying
        )
        amount = Operation.from_xdr_amount(
            operation_xdr_object.body.manageBuyOfferOp.buyAmount
        )
        price = Price.from_xdr_object(operation_xdr_object.body.manageBuyOfferOp.price)
        offer_id = operation_xdr_object.body.manageBuyOfferOp.offerID

        op = cls(
            source=source,
            selling=selling,
            buying=buying,
            amount=amount,
            price=price,
            offer_id=offer_id,
        )
        op._source_muxed = Operation.get_source_muxed_from_xdr_obj(operation_xdr_object)
        return op
